# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from datetime import datetime

import pytz
from flexmock import flexmock, flexmock_teardown

from almanach.adapters.volume_bus_adapter import VolumeBusAdapter
from integration_tests.builders import messages


class VolumeBusAdapterTest(unittest.TestCase):

    def setUp(self):
        self.controller = flexmock()
        self.retry = flexmock()
        self.volume_bus_adapter = VolumeBusAdapter(self.controller)

    def tearDown(self):
        flexmock_teardown()

    def test_deleted_volume(self):
        notification = messages.get_volume_delete_end_sample()

        self.controller.should_receive('delete_volume').once()
        self.volume_bus_adapter.on_volume_deleted(notification)

    def test_resize_volume(self):
        notification = messages.get_volume_update_end_sample()

        self.controller.should_receive('resize_volume').once()
        self.volume_bus_adapter.on_volume_resized(notification)

    def test_updated_volume(self):
        notification = messages.get_volume_update_end_sample()

        self.controller.should_receive('resize_volume').once()
        self.volume_bus_adapter.on_volume_resized(notification)

    def test_attach_volume_with_kilo_payload_and_empty_attachments(self):
        notification = messages.get_volume_attach_kilo_end_sample(
                volume_id="my-volume-id",
                timestamp=datetime(2014, 2, 14, 17, 18, 35, tzinfo=pytz.utc),
                attached_to=[]
        )

        self.controller \
            .should_receive('attach_volume') \
            .with_args("my-volume-id", "2014-02-14T17:18:36.000000Z", []) \
            .once()

        self.volume_bus_adapter.on_volume_attached(notification)

    def test_detached_volume(self):
        notification = messages.get_volume_detach_end_sample()

        (self.controller
         .should_receive('detach_volume')
         .once())

        self.volume_bus_adapter.on_volume_detached(notification)

    def test_renamed_volume_with_volume_update_end(self):
        notification = messages.get_volume_update_end_sample()

        self.controller \
            .should_receive('rename_volume') \
            .once()

        self.volume_bus_adapter.on_volume_renamed(notification)

    def test_renamed_volume_with_volume_exists(self):
        notification = messages.get_volume_exists_sample()

        self.controller.should_receive('rename_volume').once()
        self.volume_bus_adapter.on_volume_renamed(notification)

    def test_attach_volume_with_icehouse_payload(self):
        notification = messages.get_volume_attach_icehouse_end_sample(
                volume_id="my-volume-id",
                creation_timestamp=datetime(2014, 2, 14, 17, 18, 35, tzinfo=pytz.utc), attached_to="my-instance-id"
        )

        self.controller \
            .should_receive('attach_volume') \
            .with_args("my-volume-id", "2014-02-14T17:18:36.000000Z", ["my-instance-id"]) \
            .once()

        self.volume_bus_adapter.on_volume_attached(notification)

    def test_attach_volume_with_kilo_payload(self):
        notification = messages.get_volume_attach_kilo_end_sample(
                volume_id="my-volume-id",
                timestamp=datetime(2014, 2, 14, 17, 18, 35, tzinfo=pytz.utc),
                attached_to=["I1"]
        )

        self.controller \
            .should_receive('attach_volume') \
            .with_args("my-volume-id", "2014-02-14T17:18:36.000000Z", ["I1"]) \
            .once()

        self.volume_bus_adapter.on_volume_attached(notification)

    def test_get_attached_instances(self):
        self.assertEqual(["truc"], self.volume_bus_adapter._get_attached_instances({"instance_uuid": "truc"}))
        self.assertEqual([], self.volume_bus_adapter._get_attached_instances({"instance_uuid": None}))
        self.assertEqual([], self.volume_bus_adapter._get_attached_instances({}))
        self.assertEqual(
                ["a", "b"],
                self.volume_bus_adapter._get_attached_instances(
                        {"volume_attachment": [{"instance_uuid": "a"}, {"instance_uuid": "b"}]}
                )
        )
        self.assertEqual(
                ["a"],
                self.volume_bus_adapter._get_attached_instances({"volume_attachment": [{"instance_uuid": "a"}]})
        )
        self.assertEqual([], self.volume_bus_adapter._get_attached_instances({"volume_attachment": []}))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from kombu import Exchange, Queue, Producer
from almanach import config


class RetryAdapter:

    def __init__(self, connection):
        self.connection = connection
        retry_exchange = self._configure_retry_exchanges(self.connection)
        dead_exchange = self._configure_dead_exchange(self.connection)
        self._retry_producer = Producer(self.connection, exchange=retry_exchange)
        self._dead_producer = Producer(self.connection, exchange=dead_exchange)

    def publish_to_dead_letter(self, message):
        death_count = self._rejected_count(message)
        logging.info("Message has been dead {0} times".format(death_count))
        if death_count < config.rabbitmq_retry():
            logging.info("Publishing to retry queue")
            self._publish_message(self._retry_producer, message)
            logging.info("Published to retry queue")
        else:
            logging.info("Publishing to dead letter queue")
            self._publish_message(self._dead_producer, message)
            logging.info("Publishing notification to dead letter queue: {0}".format(json.dumps(message.body)))

    def _configure_retry_exchanges(self, connection):
        def declare_queues():
            channel = connection.channel()
            almanach_exchange = Exchange(name=config.rabbitmq_retry_return_exchange(),
                                         type='direct',
                                         channel=channel)
            retry_exchange = Exchange(name=config.rabbitmq_retry_exchange(),
                                      type='direct',
                                      channel=channel)
            retry_queue = Queue(name=config.rabbitmq_retry_queue(),
                                exchange=retry_exchange,
                                routing_key=config.rabbitmq_routing_key(),
                                queue_arguments=self._get_queue_arguments(),
                                channel=channel)
            almanach_queue = Queue(name=config.rabbitmq_queue(),
                                   exchange=almanach_exchange,
                                   durable=False,
                                   routing_key=config.rabbitmq_routing_key(),
                                   channel=channel)

            retry_queue.declare()
            almanach_queue.declare()

            return retry_exchange

        def error_callback(exception, interval):
            logging.error('Failed to declare queues and exchanges, retrying in %d seconds. %r' % (interval, exception))

        declare_queues = connection.ensure(connection, declare_queues, errback=error_callback,
                                           interval_start=0, interval_step=5, interval_max=30)
        return declare_queues()

    def _configure_dead_exchange(self, connection):
        def declare_dead_queue():
            channel = connection.channel()
            dead_exchange = Exchange(name=config.rabbitmq_dead_exchange(),
                                     type='direct',
                                     channel=channel)
            dead_queue = Queue(name=config.rabbitmq_dead_queue(),
                               routing_key=config.rabbitmq_routing_key(),
                               exchange=dead_exchange,
                               channel=channel)

            dead_queue.declare()

            return dead_exchange

        def error_callback(exception, interval):
            logging.error('Failed to declare dead queue and exchange, retrying in %d seconds. %r' %
                          (interval, exception))

        declare_dead_queue = connection.ensure(connection, declare_dead_queue, errback=error_callback,
                                               interval_start=0, interval_step=5, interval_max=30)
        return declare_dead_queue()

    def _get_queue_arguments(self):
        return {"x-message-ttl": self._get_time_to_live_in_seconds(),
                "x-dead-letter-exchange": config.rabbitmq_retry_return_exchange(),
                "x-dead-letter-routing-key": config.rabbitmq_routing_key()}

    def _get_time_to_live_in_seconds(self):
        return config.rabbitmq_time_to_live() * 1000

    def _rejected_count(self, message):
        if 'x-death' in message.headers:
            return len(message.headers['x-death'])
        return 0

    def _publish_message(self, producer, message):
        publish = self.connection.ensure(producer, producer.publish, errback=self._error_callback,
                                         interval_start=0, interval_step=5, interval_max=30)
        publish(message.body,
                routing_key=message.delivery_info['routing_key'],
                headers=message.headers,
                content_type=message.content_type,
                content_encoding=message.content_encoding)

    def _error_callback(self, exception, interval):
        logging.error('Failed to publish message to dead letter queue, retrying in %d seconds. %r'
                      % (interval, exception))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
import unittest
from copy import copy
from datetime import datetime, timedelta

import pytz
from dateutil.parser import parse
from hamcrest import raises, calling, assert_that, equal_to
from flexmock import flexmock, flexmock_teardown
from nose.tools import assert_raises

from almanach.common.exceptions.almanach_entity_not_found_exception import AlmanachEntityNotFoundException
from almanach.common.exceptions.multiple_entities_matching_query import MultipleEntitiesMatchingQuery
from tests.builder import a, instance, volume, volume_type
from almanach import config
from almanach.common.exceptions.date_format_exception import DateFormatException
from almanach.common.exceptions.validation_exception import InvalidAttributeException
from almanach.core.controller import Controller
from almanach.core.model import Instance, Volume

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


class ControllerTest(unittest.TestCase):
    def setUp(self):
        self.database_adapter = flexmock()

        (flexmock(config)
         .should_receive("volume_existence_threshold")
         .and_return(10))
        (flexmock(config)
         .should_receive("device_metadata_whitelist")
         .and_return(["a_metadata.to_filter"]))

        self.controller = Controller(self.database_adapter)

    def tearDown(self):
        flexmock_teardown()

    def test_instance_created(self):
        fake_instance = a(instance().with_all_dates_in_string())

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance.entity_id)
         .and_raise(KeyError)
         .once())

        expected_instance = a(instance()
                              .with_id(fake_instance.entity_id)
                              .with_project_id(fake_instance.project_id)
                              .with_metadata({"a_metadata.to_filter": "include.this"}))

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(expected_instance)
         .once())

        self.controller.create_instance(fake_instance.entity_id, fake_instance.project_id, fake_instance.start,
                                        fake_instance.flavor, fake_instance.os.os_type, fake_instance.os.distro,
                                        fake_instance.os.version, fake_instance.name, fake_instance.metadata)

    def test_resize_instance(self):
        fake_instance = a(instance())

        dates_str = "2015-10-21T16:25:00.000000Z"
        fake_instance.start = parse(dates_str)
        fake_instance.end = None
        fake_instance.last_event = parse(dates_str)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance.entity_id)
         .and_return(fake_instance)
         .once())
        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_instance.entity_id, parse(dates_str))
         .once())

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(fake_instance)
         .once())

        self.controller.resize_instance(fake_instance.entity_id, "newly_flavor", dates_str)

    def test_update_entity_closed_entity_flavor(self):
        start = datetime(2016, 3, 1, 0, 0, 0, 0, pytz.utc)
        end = datetime(2016, 3, 3, 0, 0, 0, 0, pytz.utc)
        flavor = 'a_new_flavor'
        fake_instance1 = a(instance().with_start(2016, 3, 1, 0, 0, 0).with_end(2016, 3, 2, 0, 0, 0))

        (flexmock(self.database_adapter)
         .should_receive("list_entities_by_id")
         .with_args(fake_instance1.entity_id, start, end)
         .and_return([fake_instance1])
         .twice())

        (flexmock(self.database_adapter)
         .should_receive("update_closed_entity")
         .with_args(entity=fake_instance1, data={"flavor": flavor})
         .once())

        self.controller.update_inactive_entity(
            instance_id=fake_instance1.entity_id,
            start=start,
            end=end,
            flavor=flavor,
        )

    def test_update_one_close_entity_return_multiple_entities(self):
        fake_instances = [a(instance()), a(instance())]

        (flexmock(self.database_adapter)
         .should_receive("list_entities_by_id")
         .with_args(fake_instances[0].entity_id, fake_instances[0].start, fake_instances[0].end)
         .and_return(fake_instances)
         .once())

        assert_that(
            calling(self.controller.update_inactive_entity).with_args(instance_id=fake_instances[0].entity_id,
                                                                      start=fake_instances[0].start,
                                                                      end=fake_instances[0].end,
                                                                      flavor=fake_instances[0].flavor),
            raises(MultipleEntitiesMatchingQuery)
        )

    def test_update_one_close_entity_return_no_entity(self):
        fake_instances = a(instance())

        (flexmock(self.database_adapter)
         .should_receive("list_entities_by_id")
         .with_args(fake_instances.entity_id, fake_instances.start, fake_instances.end)
         .and_return([])
         .once())

        assert_that(
            calling(self.controller.update_inactive_entity).with_args(instance_id=fake_instances.entity_id,
                                                                      start=fake_instances.start,
                                                                      end=fake_instances.end,
                                                                      flavor=fake_instances.flavor),
            raises(AlmanachEntityNotFoundException)
        )

    def test_update_active_instance_entity_with_a_new_flavor(self):
        flavor = u"my flavor name"
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, flavor=flavor)

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            flavor=flavor,
        )

    def test_update_active_instance_entity_with_a_new_name(self):
        name = u"my instance name"
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, name=name)

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            name=name,
        )

    def test_update_active_instance_entity_with_a_new_os(self):
        os = {
            "os_type": u"linux",
            "version": u"7",
            "distro": u"centos"
        }
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, os=os)

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            os=os,
        )

    def test_update_active_instance_entity_with_a_new_metadata(self):
        metadata = {
            "key": "value"
        }
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, metadata=metadata)

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            metadata=metadata,
        )

    def test_update_active_instance_entity_with_a_new_start_date(self):
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, start="2015-10-21T16:25:00.000000Z")

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            start_date="2015-10-21T16:25:00.000000Z",
        )

    def test_update_active_instance_entity_with_a_new_end_date(self):
        fake_instance1 = a(instance())
        fake_instance2 = copy(fake_instance1)
        self._expect_get_active_entity_and_update(fake_instance1, fake_instance2, end="2015-10-21T16:25:00.000000Z")

        self.controller.update_active_instance_entity(
            instance_id=fake_instance1.entity_id,
            end_date="2015-10-21T16:25:00.000000Z",
        )

    def test_instance_updated_wrong_attributes_raises_exception(self):
        fake_instance1 = a(instance())

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance1.entity_id)
         .and_return(fake_instance1)
         .never())

        assert_that(
            calling(self.controller.update_active_instance_entity).with_args(instance_id=fake_instance1.entity_id,
                                                                             wrong_attribute="this is wrong"),
            raises(InvalidAttributeException))

    def test_instance_created_but_its_an_old_event(self):
        fake_instance = a(instance()
                          .with_last_event(pytz.utc.localize(datetime(2015, 10, 21, 16, 29, 0))))

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance.entity_id)
         .and_return(fake_instance)
         .once())

        self.controller.create_instance(fake_instance.entity_id, fake_instance.project_id,
                                        '2015-10-21T16:25:00.000000Z',
                                        fake_instance.flavor, fake_instance.os.os_type, fake_instance.os.distro,
                                        fake_instance.os.version, fake_instance.name, fake_instance.metadata)

    def test_instance_created_but_find_garbage(self):
        fake_instance = a(instance().with_all_dates_in_string())

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance.entity_id)
         .and_raise(NotImplementedError)  # The db adapter found garbage in the database, we will ignore this entry
         .once())

        expected_instance = a(instance()
                              .with_id(fake_instance.entity_id)
                              .with_project_id(fake_instance.project_id)
                              .with_metadata({"a_metadata.to_filter": "include.this"}))

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(expected_instance)
         .once())

        self.controller.create_instance(fake_instance.entity_id, fake_instance.project_id, fake_instance.start,
                                        fake_instance.flavor, fake_instance.os.os_type, fake_instance.os.distro,
                                        fake_instance.os.version, fake_instance.name, fake_instance.metadata)

    def test_instance_deleted(self):
        (flexmock(self.database_adapter)
         .should_receive("has_active_entity")
         .with_args("id1")
         .and_return(True)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args("id1", parse("2015-10-21T16:25:00.000000Z"))
         .once())

        self.controller.delete_instance("id1", "2015-10-21T16:25:00.000000Z")

    def test_instance_deleted_when_entity_not_found(self):
        (flexmock(self.database_adapter)
         .should_receive("has_active_entity")
         .with_args("id1")
         .and_return(False)
         .once())

        with self.assertRaises(AlmanachEntityNotFoundException):
            self.controller.delete_instance("id1", "2015-10-21T16:25:00.000000Z")

    def test_volume_deleted(self):
        fake_volume = a(volume())

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)
        date = date + timedelta(1)
        expected_date = pytz.utc.localize(date)

        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(fake_volume.entity_id)
         .and_return(1))

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume))

        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_volume.entity_id, expected_date)
         .once())

        self.controller.delete_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%fZ"))

    def test_volume_deleted_within_volume_existance_threshold(self):
        fake_volume = a(volume())

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)
        date = date + timedelta(0, 5)

        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(fake_volume.entity_id)
         .and_return(2))

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume))

        (flexmock(self.database_adapter)
         .should_receive("delete_active_entity")
         .with_args(fake_volume.entity_id)
         .once())

        self.controller.delete_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%fZ"))

    def test_volume_deleted_within_volume_existance_threshold_but_with_only_one_entry(self):
        fake_volume = a(volume())

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)
        date = date + timedelta(0, 5)
        expected_date = pytz.utc.localize(date)

        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(fake_volume.entity_id)
         .and_return(1))

        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_volume.entity_id, expected_date)
         .once())

        self.controller.delete_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%fZ"))

    def test_list_instances(self):
        (flexmock(self.database_adapter)
         .should_receive("list_entities")
         .with_args("project_id", "start", "end", Instance.TYPE)
         .and_return(["instance1", "instance2"])
         .once())

        self.assertEqual(self.controller.list_instances("project_id", "start", "end"), ["instance1", "instance2"])

    def test_list_volumes(self):
        (flexmock(self.database_adapter)
         .should_receive("list_entities")
         .with_args("project_id", "start", "end", Volume.TYPE)
         .and_return(["volume2", "volume3"]))

        self.assertEqual(self.controller.list_volumes("project_id", "start", "end"), ["volume2", "volume3"])

    def test_list_entities(self):
        (flexmock(self.database_adapter)
         .should_receive("list_entities")
         .with_args("project_id", "start", "end")
         .and_return(["volume2", "volume3", "instance1"]))

        self.assertEqual(self.controller.list_entities(
            "project_id", "start", "end"), ["volume2", "volume3", "instance1"])

    def test_create_volume(self):
        some_volume_type = a(volume_type().with_volume_type_name("some_volume_type_name"))
        (flexmock(self.database_adapter)
         .should_receive("get_volume_type")
         .with_args(some_volume_type.volume_type_id)
         .and_return(some_volume_type)
         .once())

        some_volume = a(volume()
                        .with_volume_type(some_volume_type.volume_type_name)
                        .with_all_dates_in_string())

        expected_volume = a(volume()
                            .with_volume_type(some_volume_type.volume_type_name)
                            .with_project_id(some_volume.project_id)
                            .with_id(some_volume.entity_id))

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(expected_volume)
         .once())
        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(some_volume.entity_id)
         .and_return(None)
         .once())

        self.controller.create_volume(some_volume.entity_id, some_volume.project_id, some_volume.start,
                                      some_volume_type.volume_type_id, some_volume.size, some_volume.name,
                                      some_volume.attached_to)

    def test_create_volume_raises_bad_date_format(self):
        some_volume = a(volume())

        assert_raises(
            DateFormatException,
            self.controller.create_volume,
            some_volume.entity_id,
            some_volume.project_id,
            'bad_date_format',
            some_volume.volume_type,
            some_volume.size,
            some_volume.name,
            some_volume.attached_to
        )

    def test_create_volume_insert_none_volume_type_as_type(self):
        some_volume_type = a(volume_type().with_volume_type_id(None).with_volume_type_name(None))
        (flexmock(self.database_adapter)
         .should_receive("get_volume_type")
         .never())

        some_volume = a(volume()
                        .with_volume_type(some_volume_type.volume_type_name)
                        .with_all_dates_in_string())

        expected_volume = a(volume()
                            .with_volume_type(some_volume_type.volume_type_name)
                            .with_project_id(some_volume.project_id)
                            .with_id(some_volume.entity_id))

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(expected_volume)
         .once())
        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(some_volume.entity_id)
         .and_return(None)
         .once())

        self.controller.create_volume(some_volume.entity_id, some_volume.project_id, some_volume.start,
                                      some_volume_type.volume_type_id, some_volume.size, some_volume.name,
                                      some_volume.attached_to)

    def test_create_volume_with_invalid_volume_type(self):
        some_volume_type = a(volume_type())
        (flexmock(self.database_adapter)
         .should_receive("get_volume_type")
         .with_args(some_volume_type.volume_type_id)
         .and_raise(KeyError)
         .once())

        some_volume = a(volume()
                        .with_volume_type(some_volume_type.volume_type_name)
                        .with_all_dates_in_string())

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .never())
        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(some_volume.entity_id)
         .and_return(None)
         .once())

        with self.assertRaises(KeyError):
            self.controller.create_volume(some_volume.entity_id, some_volume.project_id, some_volume.start,
                                          some_volume_type.volume_type_id, some_volume.size, some_volume.name,
                                          some_volume.attached_to)

    def test_create_volume_but_its_an_old_event(self):
        some_volume = a(volume().with_last_event(pytz.utc.localize(datetime(2015, 10, 21, 16, 29, 0))))
        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(some_volume.entity_id)
         .and_return(some_volume)
         .once())

        self.controller.create_volume(some_volume.entity_id, some_volume.project_id, '2015-10-21T16:25:00.000000Z',
                                      some_volume.volume_type, some_volume.size, some_volume.name,
                                      some_volume.attached_to)

    def test_volume_updated(self):
        fake_volume = a(volume())
        dates_str = "2015-10-21T16:25:00.000000Z"
        fake_volume.size = "new_size"
        fake_volume.start = parse(dates_str)
        fake_volume.end = None
        fake_volume.last_event = parse(dates_str)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())
        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_volume.entity_id, parse(dates_str))
         .once())

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(fake_volume)
         .once())

        self.controller.resize_volume(fake_volume.entity_id, "new_size", dates_str)

    def test_volume_attach_with_no_existing_attachment(self):
        fake_volume = a(volume()
                        .with_no_attachment())

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(fake_volume))

        self.controller.attach_volume(
            fake_volume.entity_id,
            date.strftime("%Y-%m-%dT%H:%M:%S.%f"),
            ["new_attached_to"]
        )
        self.assertEqual(fake_volume.attached_to, ["new_attached_to"])

    def test_volume_attach_with_existing_attachments(self):
        fake_volume = a(volume()
                        .with_attached_to(["existing_attached_to"]))

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(fake_volume))

        self.controller.attach_volume(
            fake_volume.entity_id,
            date.strftime("%Y-%m-%dT%H:%M:%S.%f"),
            ["existing_attached_to", "new_attached_to"]
        )
        self.assertEqual(fake_volume.attached_to, ["existing_attached_to", "new_attached_to"])

    def test_volume_attach_after_threshold(self):
        fake_volume = a(volume())

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)
        date = date + timedelta(0, 120)
        expected_date = pytz.utc.localize(date)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_volume.entity_id, expected_date)
         .once())

        new_volume = a(volume()
                       .build_from(fake_volume)
                       .with_datetime_start(expected_date)
                       .with_no_end()
                       .with_last_event(expected_date)
                       .with_attached_to(["new_attached_to"]))

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(new_volume)
         .once())

        self.controller.attach_volume(
            fake_volume.entity_id,
            date.strftime("%Y-%m-%dT%H:%M:%S.%f"),
            ["new_attached_to"]
        )

    def test_volume_detach_with_two_attachments(self):
        fake_volume = a(volume().with_attached_to(["I1", "I2"]))

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(fake_volume))

        self.controller.detach_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%fZ"), ["I2"])
        self.assertEqual(fake_volume.attached_to, ["I2"])

    def test_volume_detach_with_one_attachments(self):
        fake_volume = a(volume().with_attached_to(["I1"]))

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(fake_volume))

        self.controller.detach_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%f"), [])
        self.assertEqual(fake_volume.attached_to, [])

    def test_volume_detach_last_attachment_after_threshold(self):
        fake_volume = a(volume().with_attached_to(["I1"]))

        date = datetime(fake_volume.start.year, fake_volume.start.month, fake_volume.start.day, fake_volume.start.hour,
                        fake_volume.start.minute, fake_volume.start.second, fake_volume.start.microsecond)
        date = date + timedelta(0, 120)
        expected_date = pytz.utc.localize(date)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .with_args(fake_volume.entity_id, expected_date)
         .once())

        new_volume = a(volume()
                       .build_from(fake_volume)
                       .with_datetime_start(expected_date)
                       .with_no_end()
                       .with_last_event(expected_date)
                       .with_no_attachment())

        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .with_args(new_volume)
         .once())

        self.controller.detach_volume(fake_volume.entity_id, date.strftime("%Y-%m-%dT%H:%M:%S.%f"), [])
        self.assertEqual(fake_volume.attached_to, [])

    def test_instance_rebuilded(self):
        i = a(instance())

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .and_return(i)
         .twice())
        (flexmock(self.database_adapter)
         .should_receive("close_active_entity")
         .once())
        (flexmock(self.database_adapter)
         .should_receive("insert_entity")
         .once())

        self.controller.rebuild_instance(
            "an_instance_id",
            "some_distro",
            "some_version",
            "some_type",
            "2015-10-21T16:25:00.000000Z"
        )
        self.controller.rebuild_instance(
            "an_instance_id",
            i.os.distro,
            i.os.version,
            i.os.os_type,
            "2015-10-21T16:25:00.000000Z"
        )

    def test_entity_exists(self):
        entity_id = "some_entity_id"
        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(entity_id=entity_id)
         .and_return(1))

        assert_that(True, equal_to(self.controller.entity_exists(entity_id)))

    def test_entity_exists_not(self):
        entity_id = "some_entity_id"
        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(entity_id=entity_id)
         .and_return(0))

        assert_that(False, equal_to(self.controller.entity_exists(entity_id)))

    def test_get_all_entities_by_id(self):
        entity_id = "some_entity_id"
        fake_entity = a(instance().with_id(entity_id))
        result = [fake_entity]

        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(entity_id=entity_id)
         .and_return(1))

        (flexmock(self.database_adapter)
         .should_receive("get_all_entities_by_id")
         .with_args(entity_id=entity_id)
         .and_return(result))

        assert_that(result, equal_to(self.controller.get_all_entities_by_id(entity_id)))

    def test_get_all_entities_by_id_when_entity_not_found(self):
        entity_id = "some_entity_id"

        (flexmock(self.database_adapter)
         .should_receive("count_entity_entries")
         .with_args(entity_id=entity_id)
         .and_return(0))

        assert_that(
            calling(self.controller.get_all_entities_by_id).with_args(entity_id),
            raises(AlmanachEntityNotFoundException)
        )

    def test_rename_volume(self):
        fake_volume = a(volume().with_display_name('old_volume_name'))

        volume_name = 'new_volume_name'

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_volume.entity_id)
         .and_return(fake_volume)
         .once())

        new_volume = a(volume().build_from(fake_volume).with_display_name(volume_name))

        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(new_volume)
         .once())

        self.controller.rename_volume(fake_volume.entity_id, volume_name)

    def test_volume_type_created(self):
        fake_volume_type = a(volume_type())

        (flexmock(self.database_adapter)
         .should_receive("insert_volume_type")
         .with_args(fake_volume_type)
         .once())

        self.controller.create_volume_type(fake_volume_type.volume_type_id, fake_volume_type.volume_type_name)

    def test_get_volume_type(self):
        some_volume = a(volume_type())
        (flexmock(self.database_adapter)
         .should_receive("get_volume_type")
         .and_return(some_volume)
         .once())

        returned_volume_type = self.controller.get_volume_type(some_volume.volume_type_id)

        self.assertEqual(some_volume, returned_volume_type)

    def test_delete_volume_type(self):
        some_volume = a(volume_type())
        (flexmock(self.database_adapter)
         .should_receive("delete_volume_type")
         .once())

        self.controller.delete_volume_type(some_volume.volume_type_id)

    def test_list_volume_types(self):
        some_volumes = [a(volume_type()), a(volume_type())]
        (flexmock(self.database_adapter)
         .should_receive("list_volume_types")
         .and_return(some_volumes)
         .once())

        self.assertEqual(len(self.controller.list_volume_types()), 2)

    def _expect_get_active_entity_and_update(self, fake_instance1, fake_instance2, **kwargs):
        for key, value in kwargs.items():
            if key in ['start', 'end']:
                value = parse(value)

            setattr(fake_instance2, key, value)

        (flexmock(self.database_adapter)
         .should_receive("get_active_entity")
         .with_args(fake_instance1.entity_id)
         .and_return(fake_instance1)
         .once())
        (flexmock(self.database_adapter)
         .should_receive("update_active_entity")
         .with_args(fake_instance2)
         .once())



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from hamcrest import has_entry, assert_that, equal_to
from retry import retry

from base_api_volume_testcase import BaseApiVolumeTestCase
from builders import messages


class CollectorVolumeDetachTest(BaseApiVolumeTestCase):
    def test_detach_volume(self):
        tenant_id = str(uuid.uuid4())
        instance_id = str(uuid.uuid4())
        volume_id = str(uuid.uuid4())

        self.rabbitMqHelper.push(message=messages.get_volume_create_end_sample(
                volume_id=volume_id, tenant_id=tenant_id, volume_type=messages.DEFAULT_VOLUME_TYPE))

        self.rabbitMqHelper.push(message=messages.get_volume_attach_kilo_end_sample(
                volume_id=volume_id, tenant_id=tenant_id, attached_to=[instance_id]))

        self.rabbitMqHelper.push(message=messages.get_volume_detach_kilo_end_sample(
                volume_id=volume_id, tenant_id=tenant_id, attached_to=[instance_id]))

        self.assert_that_volume_entity_has_instance_detached(tenant_id)

    @retry(exceptions=AssertionError, delay=1, max_delay=300)
    def assert_that_volume_entity_has_instance_detached(self, tenant_id):
        entities = self.almanachHelper.get_entities(tenant_id, "2016-01-01 00:00:00.000")
        assert_that(len(entities), equal_to(1))
        assert_that(entities[0], has_entry("end", None))
        assert_that(entities[0], has_entry("attached_to", []))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, equal_to, has_key, has_length, has_entries, is_

from almanach.common.exceptions.date_format_exception import DateFormatException
from tests.builder import instance, a
from tests.api.base_api import BaseApi, a_date_matching


class ApiInstanceTest(BaseApi):

    def test_get_instances(self):
        self.controller.should_receive('list_instances') \
            .with_args('TENANT_ID', a_date_matching("2014-01-01 00:00:00.0000"),
                       a_date_matching("2014-02-01 00:00:00.0000")) \
            .and_return([a(instance().with_id('123'))])

        code, result = self.api_get('/project/TENANT_ID/instances',
                                    query_string={
                                        'start': '2014-01-01 00:00:00.0000',
                                        'end': '2014-02-01 00:00:00.0000'
                                    },
                                    headers={'X-Auth-Token': 'some token value'})

        assert_that(code, equal_to(200))
        assert_that(result, has_length(1))
        assert_that(result[0], has_key('entity_id'))
        assert_that(result[0]['entity_id'], equal_to('123'))

    def test_update_instance_flavor_for_terminated_instance(self):
        some_new_flavor = 'some_new_flavor'
        data = dict(flavor=some_new_flavor)
        start = '2016-03-01 00:00:00.000000'
        end = '2016-03-03 00:00:00.000000'

        self.controller.should_receive('update_inactive_entity') \
            .with_args(
                instance_id="INSTANCE_ID",
                start=a_date_matching(start),
                end=a_date_matching(end),
                flavor=some_new_flavor,
        ).and_return(a(
                instance().
                with_id('INSTANCE_ID').
                with_start(2016, 03, 01, 00, 0, 00).
                with_end(2016, 03, 03, 00, 0, 00).
                with_flavor(some_new_flavor))
        )

        code, result = self.api_put(
                '/entity/instance/INSTANCE_ID',
                headers={'X-Auth-Token': 'some token value'},
                query_string={
                    'start': start,
                    'end': end,
                },
                data=data,
        )
        assert_that(code, equal_to(200))
        assert_that(result, has_key('entity_id'))
        assert_that(result, has_key('flavor'))
        assert_that(result['flavor'], is_(some_new_flavor))

    def test_successful_instance_create(self):
        self.having_config('auth_private_key', 'some token value')
        data = dict(id="INSTANCE_ID",
                    created_at="CREATED_AT",
                    name="INSTANCE_NAME",
                    flavor="A_FLAVOR",
                    os_type="AN_OS_TYPE",
                    os_distro="A_DISTRIBUTION",
                    os_version="AN_OS_VERSION")

        self.controller.should_receive('create_instance') \
            .with_args(tenant_id="PROJECT_ID",
                       instance_id=data["id"],
                       create_date=data["created_at"],
                       flavor=data['flavor'],
                       os_type=data['os_type'],
                       distro=data['os_distro'],
                       version=data['os_version'],
                       name=data['name'],
                       metadata={}) \
            .once()

        code, result = self.api_post(
                '/project/PROJECT_ID/instance',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(code, equal_to(201))

    def test_instance_create_missing_a_param_returns_bad_request_code(self):
        self.having_config('auth_private_key', 'some token value')
        data = dict(id="INSTANCE_ID",
                    created_at="CREATED_AT",
                    name="INSTANCE_NAME",
                    flavor="A_FLAVOR",
                    os_type="AN_OS_TYPE",
                    os_version="AN_OS_VERSION")

        self.controller.should_receive('create_instance') \
            .never()

        code, result = self.api_post(
                '/project/PROJECT_ID/instance',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries({"error": "The 'os_distro' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_instance_create_bad_date_format_returns_bad_request_code(self):
        self.having_config('auth_private_key', 'some token value')
        data = dict(id="INSTANCE_ID",
                    created_at="A_BAD_DATE",
                    name="INSTANCE_NAME",
                    flavor="A_FLAVOR",
                    os_type="AN_OS_TYPE",
                    os_distro="A_DISTRIBUTION",
                    os_version="AN_OS_VERSION")

        self.controller.should_receive('create_instance') \
            .with_args(tenant_id="PROJECT_ID",
                       instance_id=data["id"],
                       create_date=data["created_at"],
                       flavor=data['flavor'],
                       os_type=data['os_type'],
                       distro=data['os_distro'],
                       version=data['os_version'],
                       name=data['name'],
                       metadata={}) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_post(
                '/project/PROJECT_ID/instance',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_successful_instance_resize(self):
        data = dict(date="UPDATED_AT",
                    flavor="A_FLAVOR")

        self.controller.should_receive('resize_instance') \
            .with_args(instance_id="INSTANCE_ID",
                       flavor=data['flavor'],
                       resize_date=data['date']) \
            .once()

        code, result = self.api_put(
                '/instance/INSTANCE_ID/resize',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(code, equal_to(200))

    def test_successfull_instance_delete(self):
        data = dict(date="DELETE_DATE")

        self.controller.should_receive('delete_instance') \
            .with_args(instance_id="INSTANCE_ID",
                       delete_date=data['date']) \
            .once()

        code, result = self.api_delete('/instance/INSTANCE_ID', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(code, equal_to(202))

    def test_instance_delete_missing_a_param_returns_bad_request_code(self):
        self.controller.should_receive('delete_instance') \
            .never()

        code, result = self.api_delete(
                '/instance/INSTANCE_ID',
                data=dict(),
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries({"error": "The 'date' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_instance_delete_no_data_bad_request_code(self):
        self.controller.should_receive('delete_instance') \
            .never()

        code, result = self.api_delete('/instance/INSTANCE_ID', headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries({"error": "The request you have made must have data. None was given."}))
        assert_that(code, equal_to(400))

    def test_instance_delete_bad_date_format_returns_bad_request_code(self):
        data = dict(date="A_BAD_DATE")

        self.controller.should_receive('delete_instance') \
            .with_args(instance_id="INSTANCE_ID",
                       delete_date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_delete('/instance/INSTANCE_ID', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_instance_resize_missing_a_param_returns_bad_request_code(self):
        data = dict(date="UPDATED_AT")

        self.controller.should_receive('resize_instance') \
            .never()

        code, result = self.api_put(
                '/instance/INSTANCE_ID/resize',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries({"error": "The 'flavor' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_instance_resize_bad_date_format_returns_bad_request_code(self):
        data = dict(date="A_BAD_DATE",
                    flavor="A_FLAVOR")

        self.controller.should_receive('resize_instance') \
            .with_args(instance_id="INSTANCE_ID",
                       flavor=data['flavor'],
                       resize_date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_put(
                '/instance/INSTANCE_ID/resize',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_rebuild_instance(self):
        instance_id = 'INSTANCE_ID'
        data = {
            'distro': 'A_DISTRIBUTION',
            'version': 'A_VERSION',
            'os_type': 'AN_OS_TYPE',
            'rebuild_date': 'UPDATE_DATE',
        }
        self.controller.should_receive('rebuild_instance') \
            .with_args(
                instance_id=instance_id,
                distro=data.get('distro'),
                version=data.get('version'),
                os_type=data.get('os_type'),
                rebuild_date=data.get('rebuild_date')) \
            .once()

        code, result = self.api_put(
                '/instance/INSTANCE_ID/rebuild',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )

        assert_that(code, equal_to(200))

    def test_rebuild_instance_missing_a_param_returns_bad_request_code(self):
        data = {
            'distro': 'A_DISTRIBUTION',
            'rebuild_date': 'UPDATE_DATE',
        }

        self.controller.should_receive('rebuild_instance') \
            .never()

        code, result = self.api_put(
                '/instance/INSTANCE_ID/rebuild',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries({"error": "The 'version' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_rebuild_instance_bad_date_format_returns_bad_request_code(self):
        instance_id = 'INSTANCE_ID'
        data = {
            'distro': 'A_DISTRIBUTION',
            'version': 'A_VERSION',
            'os_type': 'AN_OS_TYPE',
            'rebuild_date': 'A_BAD_UPDATE_DATE',
        }

        self.controller.should_receive('rebuild_instance') \
            .with_args(instance_id=instance_id, **data) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_put(
                '/instance/INSTANCE_ID/rebuild',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneclient.v2_0 import client as keystone_client
from keystoneclient.v2_0.tokens import TokenManager

from almanach.auth.base_auth import BaseAuth
from almanach.common.exceptions.authentication_failure_exception import AuthenticationFailureException


class KeystoneTokenManagerFactory(object):
    def __init__(self, username, password, auth_url, tenant_name):
        self.tenant_name = tenant_name
        self.auth_url = auth_url
        self.password = password
        self.username = username

    def get_manager(self):
        return TokenManager(keystone_client.Client(
            username=self.username,
            password=self.password,
            auth_url=self.auth_url,
            tenant_name=self.tenant_name)
        )


class KeystoneAuthentication(BaseAuth):
    def __init__(self, token_manager_factory):
        self.token_manager_factory = token_manager_factory

    def validate(self, token):
        if token is None:
            raise AuthenticationFailureException("No token provided")

        try:
            self.token_manager_factory.get_manager().validate(token)
        except Exception as e:
            raise AuthenticationFailureException(e.message)

        return True



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from almanach.common.exceptions.almanach_exception import AlmanachException


class DateFormatException(AlmanachException):

    def __init__(self, message=None):
        if not message:
            message = "The provided date has an invalid format. Format should be of yyyy-mm-ddThh:mm:ss.msZ, " \
                      "ex: 2015-01-31T18:24:34.1523Z"

        super(DateFormatException, self).__init__(message)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class InstanceBusAdapter(object):
    def __init__(self, controller):
        self.controller = controller

    def handle_events(self, event_type, notification):
        if event_type == "compute.instance.create.end":
            self.on_instance_created(notification)
        elif event_type == "compute.instance.delete.end":
            self.on_instance_deleted(notification)
        elif event_type == "compute.instance.resize.confirm.end":
            self.on_instance_resized(notification)
        elif event_type == "compute.instance.rebuild.end":
            self.on_instance_rebuilt(notification)

    def on_instance_created(self, notification):
        payload = notification.get("payload")
        metadata = payload.get("metadata")

        if isinstance(metadata, list):
            metadata = {}

        self.controller.create_instance(
            payload.get("instance_id"),
            payload.get("tenant_id"),
            payload.get("created_at"),
            payload.get("instance_type"),
            payload.get("image_meta").get("os_type"),
            payload.get("image_meta").get("distro"),
            payload.get("image_meta").get("version"),
            payload.get("hostname"),
            metadata
        )

    def on_instance_deleted(self, notification):
        payload = notification.get("payload")
        date = payload.get("terminated_at")
        instance_id = payload.get("instance_id")
        self.controller.delete_instance(instance_id, date)

    def on_instance_resized(self, notification):
        payload = notification.get("payload")
        date = notification.get("timestamp")
        flavor = payload.get("instance_type")
        instance_id = payload.get("instance_id")
        self.controller.resize_instance(instance_id, flavor, date)

    def on_instance_rebuilt(self, notification):
        payload = notification.get("payload")
        date = notification.get("timestamp")
        instance_id = payload.get("instance_id")
        distro = payload.get("image_meta").get("distro")
        version = payload.get("image_meta").get("version")
        os_type = payload.get("image_meta").get("os_type")
        self.controller.rebuild_instance(instance_id, distro, version, os_type, date)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from kombu import Connection
from kombu.tests import mocks
from kombu.transport import pyamqp
from flexmock import flexmock, flexmock_teardown

from almanach import config
from almanach.adapters.retry_adapter import RetryAdapter


class BusAdapterTest(unittest.TestCase):

    def setUp(self):
        self.setup_connection_mock()
        self.setup_config_mock()

        self.retry_adapter = RetryAdapter(self.connection)

    def tearDown(self):
        flexmock_teardown()

    def setup_connection_mock(self):
        mocks.Transport.recoverable_connection_errors = pyamqp.Transport.recoverable_connection_errors
        self.connection = flexmock(Connection(transport=mocks.Transport))
        self.channel_mock = flexmock(self.connection.default_channel)
        self.connection.should_receive('channel').and_return(self.channel_mock)

    def setup_config_mock(self):
        self.config_mock = flexmock(config)
        self.config_mock.should_receive('rabbitmq_time_to_live').and_return(10)
        self.config_mock.should_receive('rabbitmq_routing_key').and_return('almanach.info')
        self.config_mock.should_receive('rabbitmq_retry_queue').and_return('almanach.retry')
        self.config_mock.should_receive('rabbitmq_dead_queue').and_return('almanach.dead')
        self.config_mock.should_receive('rabbitmq_queue').and_return('almanach.info')
        self.config_mock.should_receive('rabbitmq_retry_return_exchange').and_return('almanach')
        self.config_mock.should_receive('rabbitmq_retry_exchange').and_return('almanach.retry')
        self.config_mock.should_receive('rabbitmq_dead_exchange').and_return('almanach.dead')

    def test_declare_retry_exchanges_retries_if_it_fails(self):
        connection = flexmock(Connection(transport=mocks.Transport))
        connection.should_receive('_establish_connection').times(3)\
            .and_raise(IOError)\
            .and_raise(IOError)\
            .and_return(connection.transport.establish_connection())

        self.retry_adapter = RetryAdapter(connection)

    def test_publish_to_retry_queue_happy_path(self):
        message = MyObject
        message.headers = []
        message.body = 'omnomnom'
        message.delivery_info = {'routing_key': 42}
        message.content_type = 'xml/rapture'
        message.content_encoding = 'iso8859-1'

        self.config_mock.should_receive('rabbitmq_retry').and_return(1)
        self.expect_publish_with(message, 'almanach.retry').once()

        self.retry_adapter.publish_to_dead_letter(message)

    def test_publish_to_retry_queue_retries_if_it_fails(self):
        message = MyObject
        message.headers = {}
        message.body = 'omnomnom'
        message.delivery_info = {'routing_key': 42}
        message.content_type = 'xml/rapture'
        message.content_encoding = 'iso8859-1'

        self.config_mock.should_receive('rabbitmq_retry').and_return(2)
        self.expect_publish_with(message, 'almanach.retry').times(4)\
            .and_raise(IOError)\
            .and_raise(IOError)\
            .and_raise(IOError)\
            .and_return(message)

        self.retry_adapter.publish_to_dead_letter(message)

    def test_publish_to_dead_letter_messages_retried_more_than_twice(self):
        message = MyObject
        message.headers = {'x-death': [0, 1, 2, 3]}
        message.body = 'omnomnom'
        message.delivery_info = {'routing_key': ''}
        message.content_type = 'xml/rapture'
        message.content_encoding = 'iso8859-1'

        self.config_mock.should_receive('rabbitmq_retry').and_return(2)
        self.expect_publish_with(message, 'almanach.dead').once()

        self.retry_adapter.publish_to_dead_letter(message)

    def expect_publish_with(self, message, exchange):
        expected_message = {'body': message.body,
                            'priority': 0,
                            'content_encoding': message.content_encoding,
                            'content_type': message.content_type,
                            'headers': message.headers,
                            'properties': {'delivery_mode': 2}}

        return self.channel_mock.should_receive('basic_publish')\
            .with_args(expected_message, exchange=exchange, routing_key=message.delivery_info['routing_key'],
                       mandatory=False, immediate=False)


class MyObject(object):
    pass



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pkg_resources

from unittest import TestCase
from hamcrest import assert_that, equal_to

from almanach import config


class ConfigTest(TestCase):

    def setUp(self):
        config.read(pkg_resources.resource_filename("almanach", "resources/config/test.cfg"))

    def test_get_config_file_value(self):
        assert_that("amqp://guest:guest@localhost:5672", equal_to(config.rabbitmq_url()))

    def test_get_value_from_environment_variable(self):
        url = "amqp://openstack:openstack@hostname:5672"
        token = "my_secret"

        os.environ['RABBITMQ_URL'] = url
        os.environ['ALMANACH_AUTH_TOKEN'] = token

        assert_that(url, equal_to(config.rabbitmq_url()))
        assert_that(token, equal_to(config.auth_private_key()))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from almanach import config
from almanach.auth.mixed_auth import MixedAuthentication
from almanach.auth.keystone_auth import KeystoneAuthentication, KeystoneTokenManagerFactory
from almanach.auth.private_key_auth import PrivateKeyAuthentication


class AuthenticationAdapter(object):

    @staticmethod
    def factory():
        if config.auth_strategy() == "keystone":
            logging.info("Loading Keystone authentication backend")
            return KeystoneAuthentication(KeystoneTokenManagerFactory(
                username=config.keystone_username(),
                password=config.keystone_password(),
                auth_url=config.keystone_url(),
                tenant_name=config.keystone_tenant_name()
            ))
        elif all(auth_method in config.auth_strategy() for auth_method in ['token', 'keystone']):
            logging.info("Loading Keystone authentication backend")
            auths = [PrivateKeyAuthentication(config.auth_private_key()),
                     KeystoneAuthentication(KeystoneTokenManagerFactory(
                         username=config.keystone_username(),
                         password=config.keystone_password(),
                         auth_url=config.keystone_url(),
                         tenant_name=config.keystone_tenant_name()
                     ))]
            return MixedAuthentication(auths)
        else:
            logging.info("Loading PrivateKey authentication backend")
            return PrivateKeyAuthentication(config.auth_private_key())



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from hamcrest import equal_to, assert_that, has_entry, has_item

from base_api_volume_testcase import BaseApiVolumeTestCase
from builders import messages


class ApiVolumeTest(BaseApiVolumeTestCase):

    def test_volume_create(self):
        volume_create_query = "{url}/project/{project}/volume"
        project_id = "my_test_project_id"
        volume_id = str(uuid4())
        data = {'volume_id': volume_id,
                'attached_to': [],
                'volume_name': messages.DEFAULT_VOLUME_NAME,
                'volume_type': messages.DEFAULT_VOLUME_TYPE,
                'start': '2016-01-01T18:30:00Z',
                'size': 100}

        response = self.almanachHelper.post(url=volume_create_query, data=data, project=project_id)
        assert_that(response.status_code, equal_to(201))

        list_query = "{url}/project/{project}/volumes?start={start}&end={end}"
        response = self.almanachHelper.get(url=list_query,
                                           project=project_id,
                                           start="2016-01-01 18:29:00.000", end="2016-01-01 18:31:00.000")

        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_item({'entity_id': volume_id,
                                               'attached_to': data['attached_to'],
                                               'end': None,
                                               'name': data['volume_name'],
                                               'entity_type': 'volume',
                                               'last_event': '2016-01-01 18:30:00+00:00',
                                               'volume_type': messages.DEFAULT_VOLUME_TYPE,
                                               'start': '2016-01-01 18:30:00+00:00',
                                               'project_id': project_id,
                                               'size': data['size']}))

    def test_volume_create_bad_date_format(self):
        volume_create_query = "{url}/project/{project}/volume"
        project_id = "my_test_project_id"
        volume_id = str(uuid4())
        data = {'volume_id': volume_id,
                'attached_to': [],
                'volume_name': messages.DEFAULT_VOLUME_NAME,
                'volume_type': messages.DEFAULT_VOLUME_TYPE,
                'start': 'BAD_DATE_FORMAT',
                'size': 100}

        response = self.almanachHelper.post(url=volume_create_query, data=data, project=project_id)

        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                'error',
                'The provided date has an invalid format. Format should be of yyyy-mm-ddThh:mm:ss.msZ, '
                'ex: 2015-01-31T18:24:34.1523Z'
        ))

    def test_volume_create_missing_param(self):
        volume_create_query = "{url}/project/{project}/volume"
        project_id = "my_test_project_id"
        volume_id = str(uuid4())
        data = {'volume_id': volume_id,
                'attached_to': [],
                'volume_name': messages.DEFAULT_VOLUME_NAME,
                'volume_type': messages.DEFAULT_VOLUME_TYPE,
                'size': 100}

        response = self.almanachHelper.post(url=volume_create_query, data=data, project=project_id)

        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                "error",
                "The 'start' param is mandatory for the request you have made."
        ))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from kombu import Connection

from almanach import config
from almanach.adapters.bus_adapter import BusAdapter
from almanach.adapters.database_adapter import DatabaseAdapter
from almanach.adapters.retry_adapter import RetryAdapter
from almanach.core.controller import Controller


class AlmanachCollector(object):

    def __init__(self):
        self._controller = Controller(DatabaseAdapter())
        _connection = Connection(config.rabbitmq_url(), heartbeat=540)
        retry_adapter = RetryAdapter(_connection)
        self._busAdapter = BusAdapter(self._controller, _connection, retry_adapter)

    def run(self):
        logging.info("Listening for incoming events")
        self._busAdapter.run()



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from hamcrest import equal_to, assert_that, has_entry, has_item

from base_api_volume_testcase import BaseApiVolumeTestCase
from builders import messages


class ApiVolumeTest(BaseApiVolumeTestCase):

    def test_volume_resize(self):
        volume_create_query = "{url}/project/{project}/volume"
        project_id = "my_test_project_id"
        volume_id = str(uuid4())
        create_data = {'volume_id': volume_id,
                       'attached_to': [],
                       'volume_name': messages.DEFAULT_VOLUME_NAME,
                       'volume_type': messages.DEFAULT_VOLUME_TYPE,
                       'start': '2016-01-01T18:30:00Z',
                       'size': 100}

        response = self.almanachHelper.post(url=volume_create_query, data=create_data, project=project_id)
        assert_that(response.status_code, equal_to(201))

        resize_data = {'date': '2016-01-01T18:40:00Z',
                       'size': '150'}

        volume_resize_query = "{url}/volume/{volume_id}/resize"
        response = self.almanachHelper.put(url=volume_resize_query, data=resize_data, volume_id=volume_id)
        assert_that(response.status_code, equal_to(200))

        list_query = "{url}/project/{project}/volumes?start={start}&end={end}"
        response = self.almanachHelper.get(url=list_query,
                                           project=project_id,
                                           start="2016-01-01 18:39:00.000",
                                           end="2016-01-01 18:41:00.000")

        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_item({'entity_id': volume_id,
                                               'attached_to': create_data['attached_to'],
                                               'end': None,
                                               'name': create_data['volume_name'],
                                               'entity_type': 'volume',
                                               'last_event': '2016-01-01 18:40:00+00:00',
                                               'volume_type': messages.DEFAULT_VOLUME_TYPE,
                                               'start': '2016-01-01 18:40:00+00:00',
                                               'project_id': project_id,
                                               'size': resize_data['size']}))

    def test_volume_resize_bad_date_format(self):
        volume_resize_query = "{url}/volume/my_test_volume_id/resize"
        resize_data = {'date': 'A_BAD_DATE',
                       'size': '150'}

        response = self.almanachHelper.put(url=volume_resize_query, data=resize_data)
        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                'error',
                'The provided date has an invalid format. Format should be of yyyy-mm-ddThh:mm:ss.msZ, '
                'ex: 2015-01-31T18:24:34.1523Z'
        ))

    def test_volume_resize_missing_param(self):
        volume_resize_query = "{url}/volume/my_test_volume_id/resize"
        resize_data = {'size': '250'}

        response = self.almanachHelper.put(url=volume_resize_query, data=resize_data, instance_id="my_instance_id")
        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                "error",
                "The 'date' param is mandatory for the request you have made."
        ))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from hamcrest import assert_that, equal_to, has_item, has_entry

from base_api_testcase import BaseApiTestCase


class ApiInstanceRebuildTest(BaseApiTestCase):

    def test_instance_rebuild(self):
        instance_create_query = "{url}/project/{project}/instance"
        project_id = "my_test_project_id"
        instance_id = str(uuid4())
        create_data = {'id': instance_id,
                       'created_at': '2016-01-01T18:30:00Z',
                       'name': 'integration_test_instance_FlavorA',
                       'flavor': 'FlavorA',
                       'os_type': 'Linux',
                       'os_distro': 'Ubuntu',
                       'os_version': '12.04'}

        response = self.almanachHelper.post(url=instance_create_query, data=create_data, project=project_id)

        assert_that(response.status_code, equal_to(201))

        update_instance_rebuild_query = "{url}/instance/{instance_id}/rebuild"
        rebuild_data = {
            'distro': 'Ubuntu',
            'version': '14.04',
            'os_type': 'Linux',
            'rebuild_date': '2016-01-01T18:40:00Z'
        }

        response = self.almanachHelper.put(url=update_instance_rebuild_query, data=rebuild_data,
                                           instance_id=instance_id)

        assert_that(response.status_code, equal_to(200))

        list_query = "{url}/project/{project}/instances?start={start}&end={end}"
        response = self.almanachHelper.get(url=list_query, project=project_id,
                                           start="2016-01-01 18:39:00.000", end="2016-01-01 18:41:00.000")

        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_item({'entity_id': instance_id,
                                               'end': None,
                                               'entity_type': 'instance',
                                               'flavor': create_data['flavor'],
                                               'last_event': '2016-01-01 18:40:00+00:00',
                                               'name': create_data['name'],
                                               'os': {
                                                   'distro': create_data['os_distro'],
                                                   'os_type': create_data['os_type'],
                                                   'version': rebuild_data['version']
                                               },
                                               'project_id': project_id,
                                               'start': '2016-01-01 18:40:00+00:00',
                                               'metadata': {}}))

    def test_instance_rebuild_bad_date_format(self):
        update_instance_rebuild_query = "{url}/instance/{instance_id}/rebuild"
        instance_id = str(uuid4())
        rebuild_data = {
            'distro': 'Ubuntu',
            'version': '14.04',
            'os_type': 'Linux',
            'rebuild_date': 'A_BAD_DATE'
        }
        response = self.almanachHelper.put(url=update_instance_rebuild_query, data=rebuild_data,
                                           instance_id=instance_id)
        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                'error',
                'The provided date has an invalid format. Format should be of yyyy-mm-ddThh:mm:ss.msZ, '
                'ex: 2015-01-31T18:24:34.1523Z'
        ))

    def test_instance_rebuild_missing_param(self):
        update_instance_rebuild_query = "{url}/instance/{instance_id}/rebuild"
        instance_id = str(uuid4())
        rebuild_data = {
            'distro': 'Ubuntu',
            'os_type': 'Linux',
            'rebuild_date': 'A_BAD_DATE'
        }

        response = self.almanachHelper.put(url=update_instance_rebuild_query, data=rebuild_data,
                                           instance_id=instance_id)
        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), has_entry(
                "error",
                "The 'version' param is mandatory for the request you have made."
        ))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from hamcrest import assert_that, equal_to, has_entries

from almanach.common.exceptions.date_format_exception import DateFormatException
from tests.api.base_api import BaseApi


class ApiVolumeTest(BaseApi):

    def test_successful_volume_create(self):
        data = dict(volume_id="VOLUME_ID",
                    start="START_DATE",
                    volume_type="VOLUME_TYPE",
                    size="A_SIZE",
                    volume_name="VOLUME_NAME",
                    attached_to=["INSTANCE_ID"])

        self.controller.should_receive('create_volume') \
            .with_args(project_id="PROJECT_ID",
                       **data) \
            .once()

        code, result = self.api_post(
                '/project/PROJECT_ID/volume',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(code, equal_to(201))

    def test_volume_create_missing_a_param_returns_bad_request_code(self):
        data = dict(volume_id="VOLUME_ID",
                    start="START_DATE",
                    size="A_SIZE",
                    volume_name="VOLUME_NAME",
                    attached_to=[])

        self.controller.should_receive('create_volume') \
            .never()

        code, result = self.api_post(
                '/project/PROJECT_ID/volume',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(
                result,
                has_entries({"error": "The 'volume_type' param is mandatory for the request you have made."})
        )
        assert_that(code, equal_to(400))

    def test_volume_create_bad_date_format_returns_bad_request_code(self):
        data = dict(volume_id="VOLUME_ID",
                    start="A_BAD_DATE",
                    volume_type="VOLUME_TYPE",
                    size="A_SIZE",
                    volume_name="VOLUME_NAME",
                    attached_to=["INSTANCE_ID"])

        self.controller.should_receive('create_volume') \
            .with_args(project_id="PROJECT_ID",
                       **data) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_post(
                '/project/PROJECT_ID/volume',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_successfull_volume_delete(self):
        data = dict(date="DELETE_DATE")

        self.controller.should_receive('delete_volume') \
            .with_args(volume_id="VOLUME_ID",
                       delete_date=data['date']) \
            .once()

        code, result = self.api_delete('/volume/VOLUME_ID', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(code, equal_to(202))

    def test_volume_delete_missing_a_param_returns_bad_request_code(self):

        self.controller.should_receive('delete_volume') \
            .never()

        code, result = self.api_delete('/volume/VOLUME_ID', data=dict(), headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries({"error": "The 'date' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_volume_delete_no_data_bad_request_code(self):
        self.controller.should_receive('delete_volume') \
            .never()

        code, result = self.api_delete('/volume/VOLUME_ID', headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries({"error": "The request you have made must have data. None was given."}))
        assert_that(code, equal_to(400))

    def test_volume_delete_bad_date_format_returns_bad_request_code(self):
        data = dict(date="A_BAD_DATE")

        self.controller.should_receive('delete_volume') \
            .with_args(volume_id="VOLUME_ID",
                       delete_date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_delete('/volume/VOLUME_ID', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_successful_volume_resize(self):
        data = dict(date="UPDATED_AT",
                    size="NEW_SIZE")

        self.controller.should_receive('resize_volume') \
            .with_args(volume_id="VOLUME_ID",
                       size=data['size'],
                       update_date=data['date']) \
            .once()

        code, result = self.api_put('/volume/VOLUME_ID/resize', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(code, equal_to(200))

    def test_volume_resize_missing_a_param_returns_bad_request_code(self):
        data = dict(date="A_DATE")

        self.controller.should_receive('resize_volume') \
            .never()

        code, result = self.api_put('/volume/VOLUME_ID/resize', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries({"error": "The 'size' param is mandatory for the request you have made."}))
        assert_that(code, equal_to(400))

    def test_volume_resize_bad_date_format_returns_bad_request_code(self):
        data = dict(date="BAD_DATE",
                    size="NEW_SIZE")

        self.controller.should_receive('resize_volume') \
            .with_args(volume_id="VOLUME_ID",
                       size=data['size'],
                       update_date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_put('/volume/VOLUME_ID/resize', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_successful_volume_attach(self):
        data = dict(date="UPDATED_AT",
                    attachments=[str(uuid4())])

        self.controller.should_receive('attach_volume') \
            .with_args(volume_id="VOLUME_ID",
                       attachments=data['attachments'],
                       date=data['date']) \
            .once()

        code, result = self.api_put('/volume/VOLUME_ID/attach', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(code, equal_to(200))

    def test_volume_attach_missing_a_param_returns_bad_request_code(self):
        data = dict(date="A_DATE")

        self.controller.should_receive('attach_volume') \
            .never()

        code, result = self.api_put(
                '/volume/VOLUME_ID/attach',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(
                result,
                has_entries({"error": "The 'attachments' param is mandatory for the request you have made."})
        )
        assert_that(code, equal_to(400))

    def test_volume_attach_bad_date_format_returns_bad_request_code(self):
        data = dict(date="A_BAD_DATE",
                    attachments=[str(uuid4())])

        self.controller.should_receive('attach_volume') \
            .with_args(volume_id="VOLUME_ID",
                       attachments=data['attachments'],
                       date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_put('/volume/VOLUME_ID/attach', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))

    def test_successful_volume_detach(self):
        data = dict(date="UPDATED_AT",
                    attachments=[str(uuid4())])

        self.controller.should_receive('detach_volume') \
            .with_args(volume_id="VOLUME_ID",
                       attachments=data['attachments'],
                       date=data['date']) \
            .once()

        code, result = self.api_put('/volume/VOLUME_ID/detach', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(code, equal_to(200))

    def test_volume_detach_missing_a_param_returns_bad_request_code(self):
        data = dict(date="A_DATE")

        self.controller.should_receive('detach_volume') \
            .never()

        code, result = self.api_put('/volume/VOLUME_ID/detach', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(
                result,
                has_entries({"error": "The 'attachments' param is mandatory for the request you have made."})
        )
        assert_that(code, equal_to(400))

    def test_volume_detach_bad_date_format_returns_bad_request_code(self):
        data = dict(date="A_BAD_DATE",
                    attachments=[str(uuid4())])

        self.controller.should_receive('detach_volume') \
            .with_args(volume_id="VOLUME_ID",
                       attachments=data['attachments'],
                       date=data['date']) \
            .once() \
            .and_raise(DateFormatException)

        code, result = self.api_put('/volume/VOLUME_ID/detach', data=data, headers={'X-Auth-Token': 'some token value'})
        assert_that(result, has_entries(
                {
                    "error": "The provided date has an invalid format. "
                             "Format should be of yyyy-mm-ddThh:mm:ss.msZ, ex: 2015-01-31T18:24:34.1523Z"
                }
        ))
        assert_that(code, equal_to(400))



import unittest

from hamcrest import assert_that, calling, raises, is_

from almanach.common.exceptions.validation_exception import InvalidAttributeException
from almanach.validators.instance_validator import InstanceValidator


class InstanceValidatorTests(unittest.TestCase):
    def test_validate_update_with_invalid_attribute(self):
        instance_validator = InstanceValidator()
        payload = {"invalid attribute": ".."}
        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_name_attribute(self):
        instance_validator = InstanceValidator()
        payload = {"name": u"instance name"}

        assert_that(instance_validator.validate_update(payload), is_(payload))

    def test_validate_update_with_invalid_name_attribute(self):
        instance_validator = InstanceValidator()
        payload = {"name": 123}

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_flavor_attribute(self):
        instance_validator = InstanceValidator()
        payload = {"flavor": u"flavor"}

        assert_that(instance_validator.validate_update(payload), is_(payload))

    def test_validate_update_with_invalid_flavor_attribute(self):
        instance_validator = InstanceValidator()
        payload = {"flavor": 123}

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_start_date(self):
        instance_validator = InstanceValidator()
        payload = {"start_date": "2015-10-21T16:25:00.000000Z"}

        assert_that(instance_validator.validate_update(payload),
                    is_(payload))

    def test_validate_update_with_invalid_start_date(self):
        instance_validator = InstanceValidator()
        payload = {"start_date": "2015-10-21"}

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_end_date(self):
        instance_validator = InstanceValidator()
        payload = {"end_date": "2015-10-21T16:25:00.000000Z"}

        assert_that(instance_validator.validate_update(payload),
                    is_(payload))

    def test_validate_update_with_invalid_end_date(self):
        instance_validator = InstanceValidator()
        payload = {"end_date": "2016"}

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_os_attribute(self):
        instance_validator = InstanceValidator()
        payload = {
            "os": {
                "distro": u"centos",
                "version": u"7",
                "os_type": u"linux",
            }
        }

        assert_that(instance_validator.validate_update(payload), is_(payload))

    def test_validate_update_with_invalid_os_attribute(self):
        instance_validator = InstanceValidator()
        payload = {
            "os": {
                "distro": u"centos",
                "os_type": u"linux",
            }
        }

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))

    def test_validate_update_with_valid_metadata_attribute(self):
        instance_validator = InstanceValidator()
        payload = {
            "metadata": {
                "key": "value"
            }
        }

        assert_that(instance_validator.validate_update(payload), is_(payload))

        instance_validator = InstanceValidator()
        payload = {
            "metadata": {}
        }

        assert_that(instance_validator.validate_update(payload), is_(payload))

    def test_validate_update_with_invalid_metadata_attribute(self):
        instance_validator = InstanceValidator()
        payload = {
            "metadata": "foobar"
        }

        assert_that(calling(instance_validator.validate_update).with_args(payload),
                    raises(InvalidAttributeException))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Entity(object):
    def __init__(self, entity_id, project_id, start, end, last_event, name, entity_type):
        self.entity_id = entity_id
        self.project_id = project_id
        self.start = start
        self.end = end
        self.last_event = last_event
        self.name = name
        self.entity_type = entity_type

    def as_dict(self):
        return todict(self)

    def __eq__(self, other):
        return (other.entity_id == self.entity_id and
                other.project_id == self.project_id and
                other.start == self.start and
                other.end == self.end and
                other.last_event == self.last_event and
                other.name == self.name and
                other.entity_type == self.entity_type)


class Instance(Entity):
    TYPE = "instance"

    def __init__(self, entity_id, project_id, start, end, flavor, os, last_event, name, metadata={}, entity_type=TYPE):
        super(Instance, self).__init__(entity_id, project_id, start, end, last_event, name, entity_type)
        self.flavor = flavor
        self.metadata = metadata
        self.os = OS(**os)

    def __eq__(self, other):
        return (super(Instance, self).__eq__(other) and
                other.flavor == self.flavor and
                other.os == self.os and
                other.metadata == self.metadata)


class OS(object):
    def __init__(self, os_type, distro, version):
        self.os_type = os_type
        self.distro = distro
        self.version = version

    def __eq__(self, other):
        return (other.os_type == self.os_type and
                other.distro == self.distro and
                other.version == self.version)


class Volume(Entity):
    TYPE = "volume"

    def __init__(self, entity_id, project_id, start, end, volume_type, size, last_event, name, attached_to=None,
                 entity_type=TYPE):
        super(Volume, self).__init__(entity_id, project_id, start, end, last_event, name, entity_type)
        self.volume_type = volume_type
        self.size = size
        self.attached_to = attached_to or []

    def __eq__(self, other):
        return (super(Volume, self).__eq__(other) and
                other.volume_type == self.volume_type and
                other.size == self.size and
                other.attached_to == self.attached_to)


class VolumeType(object):
    def __init__(self, volume_type_id, volume_type_name):
        self.volume_type_id = volume_type_id
        self.volume_type_name = volume_type_name

    def __eq__(self, other):
        return other.__dict__ == self.__dict__

    def as_dict(self):
        return todict(self)


def build_entity_from_dict(entity_dict):
    if entity_dict.get("entity_type") == Instance.TYPE:
        return Instance(**entity_dict)
    elif entity_dict.get("entity_type") == Volume.TYPE:
        return Volume(**entity_dict)
    raise NotImplementedError("unsupported entity type: '%s'" % entity_dict.get("entity_type"))


def todict(obj):
    if isinstance(obj, dict):
        return obj
    elif hasattr(obj, "__iter__"):
        return [todict(v) for v in obj]
    elif hasattr(obj, "__dict__"):
        return dict([(key, todict(value))
                     for key, value in obj.__dict__.iteritems()
                     if not callable(value) and not key.startswith('_')])
    else:
        return obj



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from hamcrest import assert_that, has_entry, equal_to

from base_api_testcase import BaseApiTestCase


class ApiInstanceEntityTest(BaseApiTestCase):
    def test_update_entity_instance_with_multiple_attributes(self):
        instance_id = self._create_instance_entity()

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}",
                                           data={"start_date": "2016-04-14T18:30:00.00Z", "flavor": "FlavorB"},
                                           instance_id=instance_id,
                                           )

        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_entry("entity_id", instance_id))
        assert_that(response.json(), has_entry("start", "2016-04-14 18:30:00+00:00"))
        assert_that(response.json(), has_entry("flavor", "FlavorB"))

    def test_update_entity_instance_with_multiple_wrong_attributes(self):
        instance_id = self._create_instance_entity()

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}",
                                           data={"start_date": "2016-04-14T18:30:00.00Z", "flavor": 123, "os": 123},
                                           instance_id=instance_id,
                                           )

        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), equal_to({"error": {"flavor": "expected unicode", "os": "expected a dictionary"}}))

    def test_update_entity_instance_with_one_attribute(self):
        instance_id = self._create_instance_entity()

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}",
                                           data={"start_date": "2016-04-14T18:30:00.00Z"},
                                           instance_id=instance_id,
                                           )

        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_entry("entity_id", instance_id))
        assert_that(response.json(), has_entry("start", "2016-04-14 18:30:00+00:00"))

    def test_update_entity_instance_with_invalid_attribute(self):
        project_id = "my_test_project_id"
        instance_id = self._create_instance_entity()
        data = {
            'id': instance_id,
            'created_at': '2016-01-01T18:30:00Z',
            'name': 'integration_test_instance_FlavorA',
            'flavor': 'FlavorA',
            'os_type': 'FreeBSD',
            'os_distro': 'Stable',
            'os_version': '10',
        }

        response = self.almanachHelper.post(url="{url}/project/{project}/instance", data=data,
                                            project=project_id)
        assert_that(response.status_code, equal_to(201))

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}",
                                           data={'flavor_flavor': 'FlavorA'},
                                           instance_id=instance_id,
                                           )

        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), equal_to({"error": {"flavor_flavor": "extra keys not allowed"}}))

    def test_update_entity_instance_with_wrong_date_format(self):
        instance_id = self._create_instance_entity()

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}",
                                           data={"start_date": "my date"},
                                           instance_id=instance_id,
                                           )

        assert_that(response.status_code, equal_to(400))
        assert_that(response.json(), equal_to(
                {"error": {"start_date": "value does not match expected format %Y-%m-%dT%H:%M:%S.%fZ"}}
        ))

    def test_update_entity_change_flavor_of_closed(self):
        instance_create_query = "{url}/project/{project}/instance"
        project_id = "my_test_project_id"
        instance_id = str(uuid4())
        data = {'id': instance_id,
                'created_at': '2016-01-01T18:30:00Z',
                'name': 'integration_test_instance_FlavorA',
                'flavor': 'FlavorA',
                'os_type': 'Linux',
                'os_distro': 'Ubuntu',
                'os_version': '14.04'}

        self.almanachHelper.post(url=instance_create_query, data=data, project=project_id)
        instance_delete_query = "{url}/instance/{instance_id}"
        delete_data = {'date': '2016-01-01T18:50:00Z'}
        self.almanachHelper.delete(url=instance_delete_query, data=delete_data, instance_id=instance_id)

        response = self.almanachHelper.put(url="{url}/entity/instance/{instance_id}?start={start}&end={end}",
                                           start="2016-01-01 18:29:59.0",
                                           end="2016-01-01 18:50:00.0",
                                           data={"flavor": "FlavorB",
                                                 "end_date": "2016-01-02 18:50:00.0Z"},
                                           instance_id=instance_id,
                                           )
        assert_that(response.status_code, equal_to(200))
        assert_that(response.json(), has_entry("flavor", "FlavorB"))
        assert_that(response.json(), has_entry("end", "2016-01-02 18:50:00+00:00"))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from kombu import BrokerConnection
from kombu import Exchange
from kombu.pools import producers
from kombu.common import maybe_declare


class RabbitMqHelper(object):
    def __init__(self):
        is_container = True if os.environ.get('TEST_CONTAINER') else False
        hostname = "messaging" if is_container else "127.0.0.1"
        amqp_url = "amqp://guest:guest@{url}:{port}".format(url=hostname, port=5672)
        self.task_exchange = Exchange("almanach.info", type="topic")
        self.connection = BrokerConnection(amqp_url)

    def push(self, message):
        with producers[self.connection].acquire(block=True) as producer:
            maybe_declare(self.task_exchange, producer.channel)
            producer.publish(message, routing_key="almanach.info")



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base_api_testcase import BaseApiTestCase
from builders import messages
from helpers.mongo_helper import MongoHelper


class BaseApiVolumeTestCase(BaseApiTestCase):

    @classmethod
    def setUpClass(cls):
        cls.setup_volume_type()

    @classmethod
    def tearDownClass(cls):
        MongoHelper().drop_database()

    @classmethod
    def setup_volume_type(cls):
        cls.rabbitMqHelper.push(
                message=messages.get_volume_type_create_sample(volume_type_id=messages.DEFAULT_VOLUME_TYPE,
                                                               volume_type_name=messages.DEFAULT_VOLUME_TYPE),
        )
        cls._wait_until_volume_type_is_created(volume_type_id=messages.DEFAULT_VOLUME_TYPE)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from almanach.auth.base_auth import BaseAuth
from almanach.common.exceptions.authentication_failure_exception import AuthenticationFailureException


class MixedAuthentication(BaseAuth):
    def __init__(self, authentication_methods):
        self.authentication_methods = authentication_methods

    def validate(self, token):
        for method in self.authentication_methods:
            try:
                valid = method.validate(token)
                if valid:
                    logging.debug('Validated token with auth {0}'.format(method.__class__))
                    return True
            except AuthenticationFailureException:
                logging.debug('Failed to validate with auth {0}'.format(method.__class__))
        raise AuthenticationFailureException('No valid auth method matching token')



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, equal_to, has_key

from tests.api.base_api import BaseApi


class ApiInfoTest(BaseApi):
    def test_info(self):
        self.controller.should_receive('get_application_info').and_return({
            'info': {'version': '1.0'},
            'database': {'all_entities': 10,
                         'active_entities': 2}
        })

        code, result = self.api_get('/info')

        assert_that(code, equal_to(200))
        assert_that(result, has_key('info'))
        assert_that(result['info']['version'], equal_to('1.0'))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pymongo import MongoClient


class MongoHelper(object):

    def __init__(self):
        is_container = True if os.environ.get('TEST_CONTAINER') else False
        self.mongo = MongoClient(host="database" if is_container else "127.0.0.1")

    def drop_database(self):
        self.mongo.drop_database('almanach')



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, equal_to, has_entries, has_key, is_
from voluptuous import Invalid

from almanach.common.exceptions.validation_exception import InvalidAttributeException
from tests.builder import instance, a
from tests.api.base_api import BaseApi


class ApiEntityTest(BaseApi):

    def test_update_instance_entity_with_a_new_start_date(self):
        data = {
            "start_date": "2014-01-01 00:00:00.0000",
        }

        self.controller.should_receive('update_active_instance_entity') \
            .with_args(
                instance_id="INSTANCE_ID",
                start_date=data["start_date"],
        ).and_return(a(instance().with_id('INSTANCE_ID').with_start(2014, 01, 01, 00, 0, 00)))

        code, result = self.api_put(
                '/entity/instance/INSTANCE_ID',
                headers={'X-Auth-Token': 'some token value'},
                data=data,
        )

        assert_that(code, equal_to(200))
        assert_that(result, has_key('entity_id'))
        assert_that(result, has_key('start'))
        assert_that(result, has_key('end'))
        assert_that(result['start'], is_("2014-01-01 00:00:00+00:00"))

    def test_update_active_instance_entity_with_wrong_attribute_raise_exception(self):
        errors = [
            Invalid(message="error message1", path=["my_attribute1"]),
            Invalid(message="error message2", path=["my_attribute2"]),
        ]

        formatted_errors = {
            "my_attribute1": "error message1",
            "my_attribute2": "error message2",
        }

        instance_id = 'INSTANCE_ID'
        data = {
            'flavor': 'A_FLAVOR',
        }

        self.controller.should_receive('update_active_instance_entity') \
            .with_args(instance_id=instance_id, **data) \
            .once() \
            .and_raise(InvalidAttributeException(errors))

        code, result = self.api_put(
                '/entity/instance/INSTANCE_ID',
                data=data,
                headers={'X-Auth-Token': 'some token value'}
        )
        assert_that(result, has_entries({
            "error": formatted_errors
        }))
        assert_that(code, equal_to(400))

    def test_entity_head_with_existing_entity(self):
        entity_id = "entity_id"
        self.controller.should_receive('entity_exists') \
            .and_return(True)

        code, result = self.api_head('/entity/{id}'.format(id=entity_id), headers={'X-Auth-Token': 'some token value'})

        assert_that(code, equal_to(200))

    def test_entity_head_with_nonexistent_entity(self):
        entity_id = "entity_id"
        self.controller.should_receive('entity_exists') \
            .and_return(False)

        code, result = self.api_head('/entity/{id}'.format(id=entity_id), headers={'X-Auth-Token': 'some token value'})

        assert_that(code, equal_to(404))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from flexmock import flexmock, flexmock_teardown

from almanach.adapters.instance_bus_adapter import InstanceBusAdapter
from integration_tests.builders import messages


class InstanceBusAdapterTest(unittest.TestCase):

    def setUp(self):
        self.controller = flexmock()
        self.retry = flexmock()
        self.instance_bus_adapter = InstanceBusAdapter(self.controller)

    def tearDown(self):
        flexmock_teardown()

    def test_deleted_instance(self):
        notification = messages.get_instance_delete_end_sample()

        self.controller.should_receive('delete_instance').once()
        self.instance_bus_adapter.on_instance_deleted(notification)

    def test_instance_resized(self):
        notification = messages.get_instance_rebuild_end_sample()

        self.controller.should_receive('resize_instance').once()
        self.instance_bus_adapter.on_instance_resized(notification)

    def test_instance_rebuild(self):
        notification = messages.get_instance_rebuild_end_sample()
        self.controller \
            .should_receive("rebuild_instance") \
            .once()
        self.instance_bus_adapter.on_instance_rebuilt(notification)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import timedelta

import pytz
from dateutil import parser as date_parser

from pkg_resources import get_distribution

from almanach.common.exceptions.almanach_entity_not_found_exception import AlmanachEntityNotFoundException
from almanach.common.exceptions.date_format_exception import DateFormatException
from almanach.common.exceptions.multiple_entities_matching_query import MultipleEntitiesMatchingQuery
from almanach.core.model import Instance, Volume, VolumeType
from almanach.validators.instance_validator import InstanceValidator
from almanach import config


class Controller(object):
    def __init__(self, database_adapter):
        self.database_adapter = database_adapter
        self.metadata_whitelist = config.device_metadata_whitelist()

        self.volume_existence_threshold = timedelta(0, config.volume_existence_threshold())

    def get_application_info(self):
        return {
            "info": {"version": get_distribution("almanach").version},
            "database": {"all_entities": self.database_adapter.count_entities(),
                         "active_entities": self.database_adapter.count_active_entities()}
        }

    def create_instance(self, instance_id, tenant_id, create_date, flavor, os_type, distro, version, name, metadata):
        create_date = self._validate_and_parse_date(create_date)
        logging.info("instance %s created in project %s (flavor %s; distro %s %s %s) on %s" % (
            instance_id, tenant_id, flavor, os_type, distro, version, create_date))
        if self._fresher_entity_exists(instance_id, create_date):
            logging.warning("instance %s already exists with a more recent entry", instance_id)
            return

        filtered_metadata = self._filter_metadata_with_whitelist(metadata)

        entity = Instance(instance_id, tenant_id, create_date, None, flavor, {"os_type": os_type, "distro": distro,
                                                                              "version": version},
                          create_date, name, filtered_metadata)
        self.database_adapter.insert_entity(entity)

    def delete_instance(self, instance_id, delete_date):
        if not self.database_adapter.has_active_entity(instance_id):
            raise AlmanachEntityNotFoundException("InstanceId: {0} Not Found".format(instance_id))

        delete_date = self._validate_and_parse_date(delete_date)
        logging.info("instance %s deleted on %s" % (instance_id, delete_date))
        self.database_adapter.close_active_entity(instance_id, delete_date)

    def resize_instance(self, instance_id, flavor, resize_date):
        resize_date = self._validate_and_parse_date(resize_date)
        logging.info("instance %s resized to flavor %s on %s" % (instance_id, flavor, resize_date))
        try:
            instance = self.database_adapter.get_active_entity(instance_id)
            if flavor != instance.flavor:
                self.database_adapter.close_active_entity(instance_id, resize_date)
                instance.flavor = flavor
                instance.start = resize_date
                instance.end = None
                instance.last_event = resize_date
                self.database_adapter.insert_entity(instance)
        except KeyError as e:
            logging.error("Trying to resize an instance with id '%s' not in the database yet." % instance_id)
            raise e

    def rebuild_instance(self, instance_id, distro, version, os_type, rebuild_date):
        rebuild_date = self._validate_and_parse_date(rebuild_date)
        instance = self.database_adapter.get_active_entity(instance_id)
        logging.info("instance %s rebuilded in project %s to os %s %s %s on %s" % (instance_id, instance.project_id,
                                                                                   os_type, distro, version,
                                                                                   rebuild_date))
        if instance.os.distro != distro or instance.os.version != version:
            self.database_adapter.close_active_entity(instance_id, rebuild_date)

            instance.os.distro = distro
            instance.os.version = version
            instance.os.os_type = os_type
            instance.start = rebuild_date
            instance.end = None
            instance.last_event = rebuild_date
            self.database_adapter.insert_entity(instance)

    def update_inactive_entity(self, instance_id, start, end, **kwargs):
        inactive_entities = self.database_adapter.list_entities_by_id(instance_id, start, end)
        if len(inactive_entities) > 1:
            raise MultipleEntitiesMatchingQuery()
        if len(inactive_entities) < 1:
            raise AlmanachEntityNotFoundException("InstanceId: {0} Not Found with start".format(instance_id))
        entity = inactive_entities[0]
        entity_update = self._transform_attribute_to_match_entity_attribute(**kwargs)
        self.database_adapter.update_closed_entity(entity=entity, data=entity_update)
        start = entity_update.get('start') or start
        end = entity_update.get('end') or end
        return self.database_adapter.list_entities_by_id(instance_id, start, end)[0]

    def update_active_instance_entity(self, instance_id, **kwargs):
        try:
            InstanceValidator().validate_update(kwargs)
            instance = self.database_adapter.get_active_entity(instance_id)
            self._update_instance_object(instance, **kwargs)
            self.database_adapter.update_active_entity(instance)
            return instance
        except KeyError as e:
            logging.error("Instance '{0}' is not in the database yet.".format(instance_id))
            raise e

    def entity_exists(self, entity_id):
        return self.database_adapter.count_entity_entries(entity_id=entity_id) >= 1

    def get_all_entities_by_id(self, entity_id):
        if not self.entity_exists(entity_id=entity_id):
            raise AlmanachEntityNotFoundException("Entity not found")
        return self.database_adapter.get_all_entities_by_id(entity_id=entity_id)

    def attach_volume(self, volume_id, date, attachments):
        date = self._validate_and_parse_date(date)
        logging.info("volume %s attached to %s on %s" % (volume_id, attachments, date))
        try:
            self._volume_attach_instance(volume_id, date, attachments)
        except KeyError as e:
            logging.error("Trying to attach a volume with id '%s' not in the database yet." % volume_id)
            raise e

    def create_volume(self, volume_id, project_id, start, volume_type, size, volume_name, attached_to=None):
        start = self._validate_and_parse_date(start)
        logging.info("volume %s created in project %s to size %s on %s" % (volume_id, project_id, size, start))
        if self._fresher_entity_exists(volume_id, start):
            return

        volume_type_name = self._get_volume_type_name(volume_type)

        entity = Volume(volume_id, project_id, start, None, volume_type_name, size, start, volume_name, attached_to)
        self.database_adapter.insert_entity(entity)

    def detach_volume(self, volume_id, date, attachments):
        date = self._validate_and_parse_date(date)
        logging.info("volume %s detached on %s" % (volume_id, date))
        try:
            self._volume_detach_instance(volume_id, date, attachments)
        except KeyError as e:
            logging.error("Trying to detach a volume with id '%s' not in the database yet." % volume_id)
            raise e

    def rename_volume(self, volume_id, volume_name):
        try:
            volume = self.database_adapter.get_active_entity(volume_id)
            if volume and volume.name != volume_name:
                logging.info("volume %s renamed from %s to %s" % (volume_id, volume.name, volume_name))
                volume.name = volume_name
                self.database_adapter.update_active_entity(volume)
        except KeyError:
            logging.error("Trying to update a volume with id '%s' not in the database yet." % volume_id)

    def resize_volume(self, volume_id, size, update_date):
        update_date = self._validate_and_parse_date(update_date)
        try:
            volume = self.database_adapter.get_active_entity(volume_id)
            logging.info("volume %s updated in project %s to size %s on %s" % (volume_id, volume.project_id, size,
                                                                               update_date))
            self.database_adapter.close_active_entity(volume_id, update_date)

            volume.size = size
            volume.start = update_date
            volume.end = None
            volume.last_event = update_date
            self.database_adapter.insert_entity(volume)
        except KeyError as e:
            logging.error("Trying to update a volume with id '%s' not in the database yet." % volume_id)
            raise e

    def delete_volume(self, volume_id, delete_date):
        delete_date = self._localize_date(self._validate_and_parse_date(delete_date))
        logging.info("volume %s deleted on %s" % (volume_id, delete_date))
        try:
            if self.database_adapter.count_entity_entries(volume_id) > 1:
                volume = self.database_adapter.get_active_entity(volume_id)
                if delete_date - volume.start < self.volume_existence_threshold:
                    self.database_adapter.delete_active_entity(volume_id)
                    return
            self.database_adapter.close_active_entity(volume_id, delete_date)
        except KeyError as e:
            logging.error("Trying to delete a volume with id '%s' not in the database yet." % volume_id)
            raise e

    def create_volume_type(self, volume_type_id, volume_type_name):
        logging.info("volume type %s with name %s created" % (volume_type_id, volume_type_name))
        volume_type = VolumeType(volume_type_id, volume_type_name)
        self.database_adapter.insert_volume_type(volume_type)

    def list_entities(self, project_id, start, end):
        return self.database_adapter.list_entities(project_id, start, end)

    def list_instances(self, project_id, start, end):
        return self.database_adapter.list_entities(project_id, start, end, Instance.TYPE)

    def list_volumes(self, project_id, start, end):
        return self.database_adapter.list_entities(project_id, start, end, Volume.TYPE)

    def get_volume_type(self, type_id):
        return self.database_adapter.get_volume_type(type_id)

    def delete_volume_type(self, type_id):
        self.database_adapter.delete_volume_type(type_id)

    def list_volume_types(self):
        return self.database_adapter.list_volume_types()

    def _fresher_entity_exists(self, entity_id, date):
        try:
            entity = self.database_adapter.get_active_entity(entity_id)
            if entity and entity.last_event > date:
                return True
        except KeyError:
            pass
        except NotImplementedError:
            pass
        return False

    def _get_volume_type_name(self, volume_type_id):
        if volume_type_id is None:
            return None

        volume_type = self.database_adapter.get_volume_type(volume_type_id)
        return volume_type.volume_type_name

    def _update_instance_object(self, instance, **kwargs):
        for key, value in self._transform_attribute_to_match_entity_attribute(**kwargs).items():
            setattr(instance, key, value)
            logging.info("Updating entity for instance '{0}' with {1}={2}".format(instance.entity_id, key, value))

    def _transform_attribute_to_match_entity_attribute(self, **kwargs):
        entity = {}
        for attribute, key in dict(start="start_date", end="end_date").items():
            if kwargs.get(key):
                entity[attribute] = self._validate_and_parse_date(kwargs.get(key))

        for attribute in ["name", "flavor", "os", "metadata"]:
            if kwargs.get(attribute):
                entity[attribute] = kwargs.get(attribute)
        return entity

    def _volume_attach_instance(self, volume_id, date, attachments):
        volume = self.database_adapter.get_active_entity(volume_id)
        date = self._localize_date(date)
        volume.last_event = date
        existing_attachments = volume.attached_to
        volume.attached_to = attachments

        if existing_attachments or self._is_within_threshold(date, volume):
            self.database_adapter.update_active_entity(volume)
        else:
            self._close_volume(volume_id, volume, date)

    def _volume_detach_instance(self, volume_id, date, attachments):
        volume = self.database_adapter.get_active_entity(volume_id)
        date = self._localize_date(date)
        volume.last_event = date
        volume.attached_to = attachments

        if attachments or self._is_within_threshold(date, volume):
            self.database_adapter.update_active_entity(volume)
        else:
            self._close_volume(volume_id, volume, date)

    def _is_within_threshold(self, date, volume):
        return date - volume.start < self.volume_existence_threshold

    def _close_volume(self, volume_id, volume, date):
        self.database_adapter.close_active_entity(volume_id, date)
        volume.start = date
        volume.end = None
        self.database_adapter.insert_entity(volume)

    def _filter_metadata_with_whitelist(self, metadata):
        return {key: value for key, value in metadata.items() if key in self.metadata_whitelist}

    def _validate_and_parse_date(self, date):
        try:
            date = date_parser.parse(date)
            return self._localize_date(date)
        except TypeError:
            raise DateFormatException()

    @staticmethod
    def _localize_date(date):
        try:
            return pytz.utc.localize(date)
        except ValueError:
            return date



from hamcrest import assert_that, equal_to

from base_api_testcase import BaseApiTestCase


class ApiInstanceEntityTest(BaseApiTestCase):
    def test_head_entity_by_id_with_entity_return_200(self):
        instance_id = self._create_instance_entity()
        response = self.almanachHelper.head(url="{url}/entity/{instance_id}",
                                            instance_id=instance_id)

        assert_that(response.status_code, equal_to(200))

    def test_head_entity_by_id_without_entity_return_404(self):
        instance_id = "some_uuid"
        response = self.almanachHelper.head(url="{url}/entity/{instance_id}",
                                            instance_id=instance_id)

        assert_that(response.status_code, equal_to(404))

    def test_get_entity_by_id_with_entity_return_200(self):
        instance_id = self._create_instance_entity()
        response = self.almanachHelper.get(url="{url}/entity/{instance_id}", instance_id=instance_id)

        result = response.json()
        assert_that(response.status_code, equal_to(200))
        assert_that(len(result), equal_to(1))
        assert_that(result[0]["entity_id"], equal_to(instance_id))

    def test_get_entity_by_id_with_entity_return_404(self):
        response = self.almanachHelper.get(url="{url}/entity/{instance_id}", instance_id="foobar")

        result = response.json()
        assert_that(result, equal_to({"error": "Entity not found"}))
        assert_that(response.status_code, equal_to(404))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask

from almanach.adapters import api_route_v1 as api_route
from almanach.adapters.auth_adapter import AuthenticationAdapter
from almanach.adapters.database_adapter import DatabaseAdapter
from almanach.core.controller import Controller


class AlmanachApi(object):

    def run(self, host, port):
        api_route.controller = Controller(DatabaseAdapter())
        api_route.auth_adapter = AuthenticationAdapter().factory()

        app = Flask("almanach")
        app.register_blueprint(api_route.api)

        return app.run(host=host, port=port)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from flexmock import flexmock, flexmock_teardown
from hamcrest import raises, assert_that, calling, equal_to

from almanach.auth.keystone_auth import KeystoneAuthentication
from almanach.common.exceptions.authentication_failure_exception import AuthenticationFailureException


class KeystoneAuthenticationTest(unittest.TestCase):
    def setUp(self):
        self.token_manager_factory = flexmock()
        self.keystone_token_manager = flexmock()
        self.auth_backend = KeystoneAuthentication(self.token_manager_factory)

    def tearDown(self):
        flexmock_teardown()

    def test_with_correct_token(self):
        token = "my token"
        self.token_manager_factory.should_receive("get_manager").and_return(self.keystone_token_manager)
        self.keystone_token_manager.should_receive("validate").with_args(token)
        assert_that(self.auth_backend.validate(token), equal_to(True))

    def test_with_invalid_token(self):
        token = "bad token"
        self.token_manager_factory.should_receive("get_manager").and_return(self.keystone_token_manager)
        self.keystone_token_manager.should_receive("validate").with_args(token).and_raise(Exception)
        assert_that(calling(self.auth_backend.validate).with_args(token), raises(AuthenticationFailureException))

    def test_with_empty_token(self):
        assert_that(calling(self.auth_backend.validate).with_args(None), raises(AuthenticationFailureException))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
from datetime import datetime
from functools import wraps

import jsonpickle

from flask import Blueprint, Response, request

from werkzeug.wrappers import BaseResponse

from almanach.common.exceptions.almanach_entity_not_found_exception import AlmanachEntityNotFoundException
from almanach.common.exceptions.authentication_failure_exception import AuthenticationFailureException
from almanach.common.exceptions.multiple_entities_matching_query import MultipleEntitiesMatchingQuery
from almanach.common.exceptions.validation_exception import InvalidAttributeException
from almanach.common.exceptions.date_format_exception import DateFormatException

api = Blueprint("api", __name__)
controller = None
auth_adapter = None


def to_json(api_call):
    def encode(data):
        return jsonpickle.encode(data, unpicklable=False)

    @wraps(api_call)
    def decorator(*args, **kwargs):
        try:
            result = api_call(*args, **kwargs)
            return result if isinstance(result, BaseResponse) \
                else Response(encode(result), 200, {"Content-Type": "application/json"})
        except DateFormatException as e:
            logging.warning(e.message)
            return Response(encode({"error": e.message}), 400, {"Content-Type": "application/json"})
        except KeyError as e:
            message = "The '{param}' param is mandatory for the request you have made.".format(param=e.message)
            logging.warning(message)
            return encode({"error": message}), 400, {"Content-Type": "application/json"}
        except TypeError:
            message = "The request you have made must have data. None was given."
            logging.warning(message)
            return encode({"error": message}), 400, {"Content-Type": "application/json"}
        except InvalidAttributeException as e:
            logging.warning(e.get_error_message())
            return encode({"error": e.get_error_message()}), 400, {"Content-Type": "application/json"}
        except MultipleEntitiesMatchingQuery as e:
            logging.warning(e.message)
            return encode({"error": "Multiple entities found while updating closed"}), 400, {
                "Content-Type": "application/json"}
        except AlmanachEntityNotFoundException as e:
            logging.warning(e.message)
            return encode({"error": "Entity not found"}), 404, {"Content-Type": "application/json"}

        except Exception as e:
            logging.exception(e)
            return Response(encode({"error": e.message}), 500, {"Content-Type": "application/json"})

    return decorator


def authenticated(api_call):
    @wraps(api_call)
    def decorator(*args, **kwargs):
        try:
            auth_adapter.validate(request.headers.get('X-Auth-Token'))
            return api_call(*args, **kwargs)
        except AuthenticationFailureException as e:
            logging.error("Authentication failure: {0}".format(e.message))
            return Response('Unauthorized', 401)

    return decorator


@api.route("/info", methods=["GET"])
@to_json
def get_info():
    logging.info("Get application info")
    return controller.get_application_info()


@api.route("/project/<project_id>/instance", methods=["POST"])
@authenticated
@to_json
def create_instance(project_id):
    instance = json.loads(request.data)
    logging.info("Creating instance for tenant %s with data %s", project_id, instance)
    controller.create_instance(
        tenant_id=project_id,
        instance_id=instance['id'],
        create_date=instance['created_at'],
        flavor=instance['flavor'],
        os_type=instance['os_type'],
        distro=instance['os_distro'],
        version=instance['os_version'],
        name=instance['name'],
        metadata={}
    )

    return Response(status=201)


@api.route("/instance/<instance_id>", methods=["DELETE"])
@authenticated
@to_json
def delete_instance(instance_id):
    data = json.loads(request.data)
    logging.info("Deleting instance with id %s with data %s", instance_id, data)
    controller.delete_instance(
        instance_id=instance_id,
        delete_date=data['date']
    )

    return Response(status=202)


@api.route("/instance/<instance_id>/resize", methods=["PUT"])
@authenticated
@to_json
def resize_instance(instance_id):
    instance = json.loads(request.data)
    logging.info("Resizing instance with id %s with data %s", instance_id, instance)
    controller.resize_instance(
        instance_id=instance_id,
        resize_date=instance['date'],
        flavor=instance['flavor']
    )

    return Response(status=200)


@api.route("/instance/<instance_id>/rebuild", methods=["PUT"])
@authenticated
@to_json
def rebuild_instance(instance_id):
    instance = json.loads(request.data)
    logging.info("Rebuilding instance with id %s with data %s", instance_id, instance)
    controller.rebuild_instance(
        instance_id=instance_id,
        distro=instance['distro'],
        version=instance['version'],
        os_type=instance['os_type'],
        rebuild_date=instance['rebuild_date'],
    )

    return Response(status=200)


@api.route("/project/<project_id>/instances", methods=["GET"])
@authenticated
@to_json
def list_instances(project_id):
    start, end = get_period()
    logging.info("Listing instances between %s and %s", start, end)
    return controller.list_instances(project_id, start, end)


@api.route("/project/<project_id>/volume", methods=["POST"])
@authenticated
@to_json
def create_volume(project_id):
    volume = json.loads(request.data)
    logging.info("Creating volume for tenant %s with data %s", project_id, volume)
    controller.create_volume(
        project_id=project_id,
        volume_id=volume['volume_id'],
        start=volume['start'],
        volume_type=volume['volume_type'],
        size=volume['size'],
        volume_name=volume['volume_name'],
        attached_to=volume['attached_to']
    )

    return Response(status=201)


@api.route("/volume/<volume_id>", methods=["DELETE"])
@authenticated
@to_json
def delete_volume(volume_id):
    data = json.loads(request.data)
    logging.info("Deleting volume with id %s with data %s", volume_id, data)
    controller.delete_volume(
        volume_id=volume_id,
        delete_date=data['date']
    )

    return Response(status=202)


@api.route("/volume/<volume_id>/resize", methods=["PUT"])
@authenticated
@to_json
def resize_volume(volume_id):
    volume = json.loads(request.data)
    logging.info("Resizing volume with id %s with data %s", volume_id, volume)
    controller.resize_volume(
        volume_id=volume_id,
        size=volume['size'],
        update_date=volume['date']
    )

    return Response(status=200)


@api.route("/volume/<volume_id>/attach", methods=["PUT"])
@authenticated
@to_json
def attach_volume(volume_id):
    volume = json.loads(request.data)
    logging.info("Attaching volume with id %s with data %s", volume_id, volume)
    controller.attach_volume(
        volume_id=volume_id,
        date=volume['date'],
        attachments=volume['attachments']
    )

    return Response(status=200)


@api.route("/volume/<volume_id>/detach", methods=["PUT"])
@authenticated
@to_json
def detach_volume(volume_id):
    volume = json.loads(request.data)
    logging.info("Detaching volume with id %s with data %s", volume_id, volume)
    controller.detach_volume(
        volume_id=volume_id,
        date=volume['date'],
        attachments=volume['attachments']
    )

    return Response(status=200)


@api.route("/project/<project_id>/volumes", methods=["GET"])
@authenticated
@to_json
def list_volumes(project_id):
    start, end = get_period()
    logging.info("Listing volumes between %s and %s", start, end)
    return controller.list_volumes(project_id, start, end)


@api.route("/project/<project_id>/entities", methods=["GET"])
@authenticated
@to_json
def list_entity(project_id):
    start, end = get_period()
    logging.info("Listing entities between %s and %s", start, end)
    return controller.list_entities(project_id, start, end)


@api.route("/entity/instance/<instance_id>", methods=["PUT"])
@authenticated
@to_json
def update_instance_entity(instance_id):
    data = json.loads(request.data)
    logging.info("Updating instance entity with id %s with data %s", instance_id, data)
    if 'start' in request.args:
        start, end = get_period()
        result = controller.update_inactive_entity(instance_id=instance_id, start=start, end=end, **data)
    else:
        result = controller.update_active_instance_entity(instance_id=instance_id, **data)
    return result


@api.route("/entity/<entity_id>", methods=["HEAD"])
@authenticated
def entity_exists(entity_id):
    logging.info("Does entity with id %s exists", entity_id)
    response = Response('', 404)
    if controller.entity_exists(entity_id=entity_id):
        response = Response('', 200)
    return response


@api.route("/entity/<entity_id>", methods=["GET"])
@authenticated
@to_json
def get_entity(entity_id):
    return controller.get_all_entities_by_id(entity_id)


@api.route("/volume_types", methods=["GET"])
@authenticated
@to_json
def list_volume_types():
    logging.info("Listing volumes types")
    return controller.list_volume_types()


@api.route("/volume_type/<type_id>", methods=["GET"])
@authenticated
@to_json
def get_volume_type(type_id):
    logging.info("Get volumes type for id %s", type_id)
    return controller.get_volume_type(type_id)


@api.route("/volume_type", methods=["POST"])
@authenticated
@to_json
def create_volume_type():
    volume_type = json.loads(request.data)
    logging.info("Creating volume type with data '%s'", volume_type)
    controller.create_volume_type(
        volume_type_id=volume_type['type_id'],
        volume_type_name=volume_type['type_name']
    )
    return Response(status=201)


@api.route("/volume_type/<type_id>", methods=["DELETE"])
@authenticated
@to_json
def delete_volume_type(type_id):
    logging.info("Deleting volume type with id '%s'", type_id)
    controller.delete_volume_type(type_id)
    return Response(status=202)


def get_period():
    start = datetime.strptime(request.args["start"], "%Y-%m-%d %H:%M:%S.%f")
    if "end" not in request.args:
        end = datetime.now()
    else:
        end = datetime.strptime(request.args["end"], "%Y-%m-%d %H:%M:%S.%f")
    return start, end



from voluptuous import Schema, MultipleInvalid, Datetime, Required

from almanach.common.exceptions.validation_exception import InvalidAttributeException


class InstanceValidator(object):
    def __init__(self):
        self.schema = Schema({
            'name': unicode,
            'flavor': unicode,
            'os': {
                Required('distro'): unicode,
                Required('version'): unicode,
                Required('os_type'): unicode,
            },
            'metadata': dict,
            'start_date': Datetime(),
            'end_date': Datetime(),
        })

    def validate_update(self, payload):
        try:
            return self.schema(payload)
        except MultipleInvalid as e:
            raise InvalidAttributeException(e.errors)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytz
import uuid
from datetime import datetime
from hamcrest import has_entry, is_not, assert_that
from retry import retry

from base_api_testcase import BaseApiTestCase
from builders.messages import get_instance_delete_end_sample, get_instance_create_end_sample


class CollectorInstanceDeleteBeforeCreateTest(BaseApiTestCase):
    def test_when_instance_delete_received_before_create_instance(self):
        tenant_id = str(uuid.uuid4())
        instance_id = str(uuid.uuid4())

        self.rabbitMqHelper.push(
                get_instance_delete_end_sample(
                        instance_id=instance_id,
                        tenant_id=tenant_id,
                        deletion_timestamp=datetime(2016, 2, 1, 10, 0, 0, tzinfo=pytz.utc)
                ))

        self.rabbitMqHelper.push(
                get_instance_create_end_sample(
                        instance_id=instance_id,
                        tenant_id=tenant_id,
                        creation_timestamp=datetime(2016, 2, 1, 9, 0, 0, tzinfo=pytz.utc)
                ))

        self.assert_instance_delete_received_before_instance_create(tenant_id)

    @retry(exceptions=AssertionError, delay=1, max_delay=300)
    def assert_instance_delete_received_before_instance_create(self, tenant_id):
        assert_that(self.almanachHelper.get_entities(tenant_id, "2016-01-01 00:00:00.000"),
                    is_not(has_entry("end", None)))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ConfigParser
import os
import os.path as os_path

from almanach.common.exceptions.almanach_exception import AlmanachException

configuration = ConfigParser.RawConfigParser()


def read(filename):
    if not os_path.isfile(filename):
        raise AlmanachException("Config file '{0}' not found".format(filename))

    print("Loading configuration file {0}".format(filename))
    configuration.read(filename)


def get(section, option, default=None):
    value = os.environ.get(section + "_" + option.upper())

    if value:
        return value

    try:
        return configuration.get(section, option)
    except:
        return default


def volume_existence_threshold():
    return int(get("ALMANACH", "volume_existence_threshold"))


def auth_strategy():
    return get("ALMANACH", "auth_strategy", "private_key")


def auth_private_key():
    return get("ALMANACH", "auth_token")


def keystone_username():
    return get("KEYSTONE", "username")


def keystone_password():
    return get("KEYSTONE", "password")


def keystone_url():
    return get("KEYSTONE", "auth_url")


def keystone_tenant_name():
    return get("KEYSTONE", "tenant_name")


def device_metadata_whitelist():
    return get("ALMANACH", "device_metadata_whitelist").split(',')


def mongodb_url():
    return get("MONGODB", "url", default=None)


def mongodb_database():
    return get("MONGODB", "database", default="almanach")


def mongodb_indexes():
    return get('MONGODB', 'indexes').split(',')


def rabbitmq_url():
    return get("RABBITMQ", "url", default=None)


def rabbitmq_queue():
    return get("RABBITMQ", "queue", default=None)


def rabbitmq_exchange():
    return get("RABBITMQ", "exchange", default=None)


def rabbitmq_routing_key():
    return get("RABBITMQ", "routing.key", default=None)


def rabbitmq_retry():
    return int(get("RABBITMQ", "retry.maximum", default=None))


def rabbitmq_retry_exchange():
    return get("RABBITMQ", "retry.exchange", default=None)


def rabbitmq_retry_return_exchange():
    return get("RABBITMQ", "retry.return.exchange", default=None)


def rabbitmq_retry_queue():
    return get("RABBITMQ", "retry.queue", default=None)


def rabbitmq_dead_queue():
    return get("RABBITMQ", "dead.queue", default=None)


def rabbitmq_dead_exchange():
    return get("RABBITMQ", "dead.exchange", default=None)


def rabbitmq_time_to_live():
    return int(get("RABBITMQ", "retry.time.to.live", default=None))


def _read_file(filename):
    f = open(filename, "r")
    content = f.read()
    f.close()
    return content
